from __future__ import annotations

import dataclasses
import enum
import inspect
import datetime
import ipaddress
import pathlib
import re
import uuid
from collections import deque, abc
from decimal import Decimal
from typing import (
    Mapping,
    Type,
    Union,
    Optional,
    List,
    Any,
    Dict,
    Hashable,
    cast,
    Set,
    ClassVar,
    Deque,
    Tuple,
)

from typic.checks import (
    isuniontype,
    isoptionaltype,
    isbuiltintype,
    isconstrained,
    isforwardref,
    istypeddict,
    isbuiltinsubtype,
    isnamedtuple,
    should_unwrap,
    isclassvartype,
    isenumtype,
)
from typic.compat import Literal, lru_cache
from typic.types import dsn, email, frozendict, path, secret, url
from typic.util import (
    origin,
    get_args,
    get_tag_for_types,
    cached_signature,
    cached_type_hints,
    get_name,
    TypeMap,
    empty,
)
from .array import (
    Array,
    FrozenSetConstraints,
    ListConstraints,
    SetContraints,
    TupleConstraints,
)
from .common import (
    MultiConstraints,
    TypeConstraints,
    EnumConstraints,
    VT,
    ConstraintsProtocolT,
    DelayedConstraints,
    ForwardDelayedConstraints,
    LiteralConstraints,
)
from .mapping import (
    MappingConstraints,
    DictConstraints,
    ObjectConstraints,
    TypedDictConstraints,
)
from .number import (
    IntContraints,
    FloatContraints,
    DecimalContraints,
    NumberT,
)
from .text import BytesConstraints, StrConstraints


@lru_cache(maxsize=None)
def get_constraints(
    t: Type[VT],
    *,
    nullable: bool = False,
    name: str = None,
    cls: Optional[Type] = ...,  # type: ignore
) -> ConstraintsProtocolT[VT]:
    while should_unwrap(t):
        nullable = nullable or isoptionaltype(t)
        t = get_args(t)[0]
    if t is cls or t in __stack:
        dc = DelayedConstraints(
            t, nullable=nullable, name=name, factory=get_constraints
        )
        return cast(ConstraintsProtocolT, dc)
    if isforwardref(t):
        if cls is ...:  # pragma: nocover
            raise TypeError(
                f"Cannot build constraints for {t} without an enclosing class."
            )
        fdc = ForwardDelayedConstraints(
            t,  # type: ignore
            cls.__module__,
            localns=getattr(cls, "__dict__", {}).copy(),
            nullable=nullable,
            name=name,
            factory=get_constraints,
        )
        return cast(ConstraintsProtocolT, fdc)
    if isconstrained(t):
        c: ConstraintsProtocolT = t.__constraints__  # type: ignore
        if (c.name, c.nullable) != (name, nullable):
            return dataclasses.replace(c, name=name, nullable=nullable)
        return c
    if isenumtype(t):
        ec = _from_enum_type(t, nullable=nullable, name=name)  # type: ignore
        return cast(ConstraintsProtocolT, ec)
    if isnamedtuple(t) or istypeddict(t):
        handler = _from_class
    else:
        ot = origin(t)
        if ot in {type, abc.Callable}:
            handler = _from_strict_type  # type: ignore
            t = ot
        else:
            handler = _CONSTRAINT_BUILDER_HANDLERS.get_by_parent(ot, _from_class)  # type: ignore

    __stack.add(t)
    c = handler(t, nullable=nullable, name=name, cls=cls)
    __stack.clear()
    return c


__stack: Set[Type] = set()


ConstraintsT = Union[
    BytesConstraints,
    DecimalContraints,
    DelayedConstraints,
    DictConstraints,
    EnumConstraints,
    FloatContraints,
    ForwardDelayedConstraints,
    FrozenSetConstraints,
    IntContraints,
    ListConstraints,
    MappingConstraints,
    MultiConstraints,
    ObjectConstraints,
    SetContraints,
    StrConstraints,
    TupleConstraints,
    TypeConstraints,
]

_ARRAY_CONSTRAINTS_BY_TYPE = TypeMap(
    {
        set: SetContraints,
        list: ListConstraints,
        tuple: TupleConstraints,
        frozenset: FrozenSetConstraints,
    }
)
ArrayConstraintsT = Union[
    SetContraints, ListConstraints, TupleConstraints, FrozenSetConstraints
]


def _resolve_args(
    *args, cls: Type = None, nullable: bool = False, multi: bool = True
) -> Optional[Union[ConstraintsProtocolT, Tuple[ConstraintsProtocolT, ...]]]:
    largs: Deque = deque(args)
    items: List[ConstraintsProtocolT] = []
    while largs:
        arg = largs.popleft()
        if arg in {Any, Ellipsis}:
            continue
        if isuniontype(arg):
            c = _from_union(arg, cls=cls, nullable=nullable)
            # just extend the outer multi constraints if that's what we're building
            if isinstance(c, MultiConstraints) and multi:
                items.extend(c.constraints)
            else:
                items.append(c)
            continue
        items.append(get_constraints(arg, cls=cls, nullable=nullable))
    if len(items) == 1:
        return items[0]
    if multi:
        return cast(ConstraintsProtocolT, MultiConstraints((*items,)))
    return (*items,)


def _from_array_type(
    t: Type[Array], *, nullable: bool = False, name: str = None, cls: Type = None
) -> ArrayConstraintsT:
    args = get_args(t)
    constr_class = cast(
        Type[ArrayConstraintsT], _ARRAY_CONSTRAINTS_BY_TYPE.get_by_parent(origin(t))
    )
    # If we don't have args, then return a naive constraint
    if not args:
        return constr_class(nullable=nullable, name=name)

    if constr_class is TupleConstraints and ... not in args:
        items = _resolve_args(*args, cls=cls, nullable=nullable, multi=False)
        return constr_class(nullable=nullable, values=items, name=name)  # type: ignore

    items = _resolve_args(*args, cls=cls, nullable=nullable, multi=True)
    return constr_class(nullable=nullable, values=items, name=name)  # type: ignore


def _from_mapping_type(
    t: Type[Mapping], *, nullable: bool = False, name: str = None, cls: Type = None
) -> Union[MappingConstraints, DictConstraints]:
    if isbuiltintype(t):
        return DictConstraints(nullable=nullable, name=name)
    base = getattr(t, "__origin__", t)
    constr_class: Union[Type[MappingConstraints], Type[DictConstraints]]
    constr_class = MappingConstraints
    if base is dict:
        constr_class = DictConstraints
    args = get_args(t)
    if not args:
        return constr_class(nullable=nullable, name=name)
    key_arg, value_arg = args
    key_items, value_items = (
        _resolve_args(key_arg, cls=cls),
        _resolve_args(value_arg, cls=cls),
    )
    return constr_class(
        keys=key_items, values=value_items, nullable=nullable, name=name  # type: ignore
    )


SimpleT = Union[NumberT, str, bytes]
SimpleConstraintsT = Union[
    IntContraints, FloatContraints, DecimalContraints, StrConstraints, BytesConstraints
]
_SIMPLE_CONSTRAINTS = TypeMap(
    {
        IntContraints.type: IntContraints,
        FloatContraints.type: FloatContraints,
        DecimalContraints.type: DecimalContraints,
        StrConstraints.type: StrConstraints,
        BytesConstraints.type: BytesConstraints,
    }
)


def _from_simple_type(
    t: Type[SimpleT], *, nullable: bool = False, name: str = None, cls: Type = None
) -> SimpleConstraintsT:
    constr_class = cast(
        Type[SimpleConstraintsT], _SIMPLE_CONSTRAINTS.get_by_parent(origin(t))
    )
    return constr_class(nullable=nullable, name=name)


def _resolve_params(
    cls: Type,
    **param: inspect.Parameter,
) -> Mapping[str, ConstraintsProtocolT]:
    items: Dict[str, ConstraintsProtocolT] = {}
    while param:
        name, p = param.popitem()
        anno = p.annotation
        nullable = p.default in (None, Ellipsis) or isoptionaltype(anno)
        if anno in {Any, Ellipsis, p.empty}:
            continue
        if isuniontype(anno) and not isforwardref(anno):
            items[name] = _from_union(anno, nullable=nullable, name=name, cls=cls)
            continue
        else:
            items[name] = get_constraints(anno, nullable=nullable, name=name, cls=cls)
    return items


def _from_strict_type(
    t: Type[VT], *, nullable: bool = False, name: str = None, cls: Type = None
) -> TypeConstraints:
    return TypeConstraints(t, nullable=nullable, name=name)


def _from_enum_type(
    t: Type[enum.Enum], *, nullable: bool = False, name: str = None, cls: Type = None
) -> EnumConstraints:
    return EnumConstraints(t, nullable=nullable, name=name)


def _from_literal(
    t: Type[VT], *, nullable: bool = False, name: str = None, cls: Type = None
) -> LiteralConstraints:
    return LiteralConstraints(t, nullable=nullable, name=name)


def _from_union(
    t: Type[VT], *, nullable: bool = False, name: str = None, cls: Type = None
) -> ConstraintsProtocolT:
    _nullable: bool = isoptionaltype(t)
    nullable = nullable or _nullable
    _args = get_args(t)[:-1] if _nullable else get_args(t)
    if len(_args) == 1:
        return get_constraints(_args[0], nullable=nullable, name=name, cls=cls)
    c = MultiConstraints(
        (*(get_constraints(a, nullable=nullable, cls=cls) for a in _args),),
        name=name,
        tag=get_tag_for_types(_args),
    )
    return cast(ConstraintsProtocolT, c)


def _from_class(
    t: Type[VT], *, nullable: bool = False, name: str = None, cls: Type = None
) -> ConstraintsProtocolT[VT]:
    if not istypeddict(t) and not isnamedtuple(t) and isbuiltinsubtype(t):
        return cast(
            ConstraintsProtocolT, _from_strict_type(t, nullable=nullable, name=name)
        )
    try:
        params: Dict[str, inspect.Parameter] = {**cached_signature(t).parameters}
        hints = cached_type_hints(t)
        for x in hints.keys() & params.keys():
            p = params[x]
            params[x] = inspect.Parameter(
                p.name, p.kind, default=p.default, annotation=hints[x]
            )
        for x in hints.keys() - params.keys():
            hint = hints[x]
            if not isclassvartype(hint):
                continue
            # Hack in the classvars as "parameters" to allow for validation.
            default = getattr(t, x, empty)
            args = get_args(hint)
            if not args:
                hint = ClassVar[default.__class__]  # type: ignore
            params[x] = inspect.Parameter(
                x, inspect.Parameter.KEYWORD_ONLY, default=default, annotation=hint
            )
    except (ValueError, TypeError):
        return cast(
            ConstraintsProtocolT, _from_strict_type(t, nullable=nullable, name=name)
        )
    name = name or get_name(t)
    items: Optional[frozendict.FrozenDict[Hashable, ConstraintsT]] = (
        frozendict.FrozenDict(_resolve_params(t, **params)) or None
    )
    required = frozenset(
        (
            pname
            for pname, p in params.items()
            if (
                p.kind not in {p.VAR_POSITIONAL, p.VAR_KEYWORD} and p.default is p.empty
            )
        )
    )
    has_varargs = any(
        p.kind in {p.VAR_KEYWORD, p.VAR_POSITIONAL} for p in params.values()
    )
    kwargs = {
        "type": t,
        "nullable": nullable,
        "name": name,
        "required_keys": required,
        "items": items,
        "total": not has_varargs,
    }
    cls = ObjectConstraints
    if istypeddict(t):
        cls = TypedDictConstraints
        kwargs.update(type=dict, ttype=t, total=getattr(t, "__total__", bool(required)))
    c = cls(**kwargs)  # type: ignore
    return cast(ConstraintsProtocolT, c)


_CONSTRAINT_BUILDER_HANDLERS = TypeMap(
    {
        set: _from_array_type,
        frozenset: _from_array_type,
        list: _from_array_type,
        tuple: _from_array_type,
        dict: _from_mapping_type,  # type: ignore
        int: _from_simple_type,
        float: _from_simple_type,
        Decimal: _from_simple_type,
        str: _from_simple_type,
        bytes: _from_simple_type,
        bool: _from_strict_type,
        datetime.datetime: _from_strict_type,
        datetime.date: _from_strict_type,
        datetime.time: _from_strict_type,
        url.NetworkAddress: _from_strict_type,
        url.URL: _from_strict_type,
        url.AbsoluteURL: _from_strict_type,
        url.RelativeURL: _from_strict_type,
        dsn.DSN: _from_strict_type,
        pathlib.Path: _from_strict_type,
        path.FilePath: _from_strict_type,
        path.DirectoryPath: _from_strict_type,
        path.PathType: _from_strict_type,
        url.HostName: _from_strict_type,
        email.Email: _from_strict_type,
        secret.SecretStr: _from_strict_type,
        secret.SecretBytes: _from_strict_type,
        uuid.UUID: _from_strict_type,
        re.Pattern: _from_strict_type,  # type: ignore
        ipaddress.IPv4Address: _from_strict_type,
        ipaddress.IPv6Address: _from_strict_type,
        Union: _from_union,  # type: ignore
        Literal: _from_literal,  # type: ignore
    }
)



from __future__ import annotations

import dataclasses
import datetime
import functools
import inspect
import pathlib
import re
import uuid
from collections import deque, defaultdict, abc
from operator import attrgetter
from typing import (
    Mapping,
    Any,
    Union,
    Callable,
    Type,
    Dict,
    Tuple,
    List,
    Pattern,
    Match,
    cast,
    TYPE_CHECKING,
    Optional,
    Set,
    NoReturn,
)

from pendulum import parse as dateparse, DateTime, instance

from typic import checks, gen, constraints as const
from typic.strict import STRICT_MODE
from typic.util import (
    safe_eval,
    cached_issubclass,
    cached_signature,
    get_defname,
    get_tag_for_types,
    get_unique_name,
    get_name,
    slotted,
)
from typic.common import DEFAULT_ENCODING, VAR_POSITIONAL, VAR_KEYWORD, ObjectT
from typic.compat import TypeGuard, Literal
from .common import (
    DeserializerT,
    DeserializerRegistryT,
    SerdeConfig,
    Annotation,
    DelayedAnnotation,
    ForwardDelayedAnnotation,
    AnnotationT,
)

if TYPE_CHECKING:  # pragma: nocover
    from .resolver import Resolver

_ORIG_SETTER_NAME = "__setattr_original__"
_origsettergetter = attrgetter(_ORIG_SETTER_NAME)
_TYPIC_ANNOS_NAME = "__typic_annotations__"
_annosgetter = attrgetter(_TYPIC_ANNOS_NAME)
_TOO_MANY_POS = "too many positional arguments"
_VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL
_VAR_KEYWORD = inspect.Parameter.VAR_KEYWORD
_KEYWORD_ONLY = inspect.Parameter.KEYWORD_ONLY
_POSITIONAL_ONLY = inspect.Parameter.POSITIONAL_ONLY
_POSITIONAL_OR_KEYWORD = inspect.Parameter.POSITIONAL_OR_KEYWORD
_KWD_KINDS = {_VAR_KEYWORD, _KEYWORD_ONLY}
_POS_KINDS = {_VAR_POSITIONAL, _POSITIONAL_ONLY}
_empty = inspect.Signature.empty
_RETURN_KEY = "return"
_SELF_NAME = "self"
_TO_RESOLVE: List[Union[Type, Callable]] = []
_SCHEMA_NAME = "__json_schema__"


class DesFactory:
    """A callable class for ``des``erialzing values.

    Checks for:

            - builtin types
            - :py:mod:`typing` type annotations
            - :py:class:`datetime.date`
            - :py:class:`datetime.datetime`
            - :py:class:`typing.TypedDict`
            - :py:class:`typing.NamedTuple`
            - :py:func:`collections.namedtuple`
            - User-defined classes (limited)

    Examples
    --------
    >>> import typic
    >>> typic.transmute(bytes, "foo")
    b'foo'
    >>> typic.transmute(dict, "{'foo': 'bar'}")
    {'foo': 'bar'}
    """

    STRICT = STRICT_MODE
    DEFAULT_BYTE_ENCODING = "utf-8"
    UNRESOLVABLE = frozenset(
        (
            Any,
            Match,
            re.Match,  # type: ignore
            type(None),
            _empty,
            Callable,
            abc.Callable,
        )
    )
    VNAME = "val"
    VTYPE = "vtype"
    __DES_CACHE: Dict[str, Tuple[DeserializerT, "const.ValidatorT"]] = {}
    __USER_DESS: DeserializerRegistryT = deque()

    def __init__(self, resolver: Resolver):
        self.resolver = resolver

    def register(self, deserializer: DeserializerT, check: DeserializerT):
        """Register a user-defined coercer.

        In the rare case where typic can't figure out how to coerce your annotation
        correctly, a custom coercer may be registered alongside a check function which
        returns a simple boolean indicating whether this is the correct coercer for an
        annotation.
        """
        self.__USER_DESS.appendleft((check, deserializer))

    def _set_checks(self, func: gen.Block, anno_name: str, annotation: Annotation):
        _ctx = {}
        # run a safe eval if input is text and anno isn't
        if inspect.isclass(annotation.resolved_origin) and issubclass(
            annotation.resolved_origin, (str, bytes)
        ):
            self._add_vtype(func)
        else:
            self._add_eval(func)
        # Equality checks for defaults and optionals
        custom_equality = hasattr(annotation.resolved_origin, "equals")
        if custom_equality and (annotation.optional or annotation.has_default):
            func.l(f"custom_equality = hasattr({self.VNAME}, 'equals')")
        null = ""
        if annotation.optional:
            null = f"{self.VNAME} in {self.resolver.OPTIONALS}"
            if custom_equality:
                null = (
                    f"(any({self.VNAME}.equals(o) for o in {self.resolver.OPTIONALS}) "
                    "if custom_equality "
                    f"else {null})"
                )
        eq = ""
        if (
            annotation.has_default
            and annotation.parameter.default not in self.resolver.OPTIONALS
        ):
            eq = f"{self.VNAME} == __default"
            if custom_equality:
                if hasattr(annotation.parameter.default, "equals"):
                    eq = f"__default.equals({self.VNAME})"
                eq = f"{self.VNAME}.equals(__default) if custom_equality else {eq}"
            _ctx["__default"] = annotation.parameter.default
        if eq or null:
            # Add a type-check for anything that isn't a builtin.
            if eq and not checks.isbuiltintype(annotation.resolved_origin):
                eq = f"{self.VTYPE} is {anno_name} and {eq}"
            check = " or ".join(c for c in (null, eq) if c)
            with func.b(f"if {check}:", **_ctx) as b:  # type: ignore
                b.l(f"return {self.VNAME}")

    @staticmethod
    def _get_name(
        annotation: Annotation, constr: Optional[const.ConstraintsProtocolT]
    ) -> str:
        return get_defname("deserializer", (annotation, constr))

    def _build_date_des(self, context: BuildContext):
        func, annotation, anno_name = (
            context.func,
            context.annotation,
            context.anno_name,
        )
        origin = annotation.resolved_origin
        # From an int
        with func.b(f"if isinstance({self.VNAME}, (int, float)):") as b:
            b.l(f"{self.VNAME} = {anno_name}.fromtimestamp({self.VNAME})")
        # From a string
        with func.b(f"elif isinstance({self.VNAME}, (str, bytes)):") as b:
            line = f"{self.VNAME} = dateparse({self.VNAME})"
            b.l(line, dateparse=dateparse)
        if issubclass(origin, datetime.datetime):
            with func.b(
                f"if isinstance({self.VNAME}, datetime):", datetime=datetime.datetime
            ) as b:
                # Use pendulum's helper if possible.
                if origin is DateTime:
                    b.l(f"{self.VNAME} = instance({self.VNAME})", instance=instance)
                else:
                    b.l(
                        f"{self.VNAME} = "
                        f"{anno_name}("
                        f"{self.VNAME}.year, "
                        f"{self.VNAME}.month, "
                        f"{self.VNAME}.day, "
                        f"{self.VNAME}.hour, "
                        f"{self.VNAME}.minute, "
                        f"{self.VNAME}.second, "
                        f"{self.VNAME}.microsecond, "
                        f"{self.VNAME}.tzinfo"
                        f")",
                    )
            with func.b(
                f"elif isinstance({self.VNAME}, date):", date=datetime.date
            ) as b:
                b.l(
                    f"{self.VNAME} = "
                    f"{anno_name}("
                    f"{self.VNAME}.year, "
                    f"{self.VNAME}.month, "
                    f"{self.VNAME}.day"
                    f")",
                )
        elif issubclass(origin, datetime.date):
            with func.b(
                f"if isinstance({self.VNAME}, datetime):", datetime=datetime.datetime
            ) as b:
                b.l(f"{self.VNAME} = {self.VNAME}.date()")
        with func.b(f"elif isinstance({self.VNAME}, (int, float)):") as b:
            b.l(f"{self.VNAME} = {anno_name}.fromtimestamp({self.VNAME})")
        with func.b(f"elif isinstance({self.VNAME}, (str, bytes)):") as b:
            line = f"{self.VNAME} = dateparse({self.VNAME}, exact=True)"
            b.l(line, dateparse=dateparse)

    def _build_time_des(self, context: BuildContext):
        func, anno_name = context.func, context.anno_name
        # From an int
        with func.b(f"if isinstance({self.VNAME}, (int, float)):") as b:
            b.l(f"{self.VNAME} = {anno_name}(int({self.VNAME}))")
        # From a string
        with func.b(f"elif isinstance({self.VNAME}, (str, bytes)):") as b:
            line = f"{self.VNAME} = dateparse({self.VNAME}, exact=True)"
            b.l(line, dateparse=dateparse)
        # From a datetime
        with func.b(
            f"if isinstance({self.VNAME}, datetime):", datetime=datetime.datetime
        ) as b:
            b.l(f"{self.VNAME} = {self.VNAME}.time()")
        # From a date
        with func.b(f"if isinstance({self.VNAME}, date):", date=datetime.date) as b:
            b.l(f"{self.VNAME} = {anno_name}(0)")

    def _build_timedelta_des(self, context: BuildContext):
        func, anno_name = context.func, context.anno_name
        # From an int
        with func.b(f"if isinstance({self.VNAME}, (int, float)):") as b:
            b.l(f"{self.VNAME} = {anno_name}(int({self.VNAME}))")
        # From a string
        with func.b(f"elif isinstance({self.VNAME}, (str, bytes)):") as b:
            line = f"{self.VNAME} = dateparse({self.VNAME}, exact=True)"
            b.l(line, dateparse=dateparse)

    def _build_uuid_des(self, context: BuildContext):
        func, anno_name = context.func, context.anno_name
        self._add_type_check(func, anno_name)
        with func.b(f"if issubclass({self.VTYPE}, UUID):", UUID=uuid.UUID) as b:
            b.l(f"{self.VNAME} = {anno_name}(int={self.VNAME}.int)")

        with func.b(f"elif isinstance({self.VNAME}, str):") as b:
            b.l(f"{self.VNAME} = {anno_name}({self.VNAME})")

        with func.b(f"elif isinstance({self.VNAME}, bytes):") as b:
            b.l(f"{self.VNAME} = {anno_name}(bytes={self.VNAME})")

        with func.b(f"elif isinstance({self.VNAME}, int):") as b:
            b.l(f"{self.VNAME} = {anno_name}(int={self.VNAME})")

        with func.b(f"elif isinstance({self.VNAME}, tuple):") as b:
            b.l(f"{self.VNAME} = {anno_name}(fields={self.VNAME})")

    def _add_eval(self, func: gen.Block):
        func.l(
            f"_, {self.VNAME} = __eval({self.VNAME}) "
            f"if isinstance({self.VNAME}, (str, bytes)) "
            f"else (False, {self.VNAME})",
            __eval=safe_eval,
        )
        self._add_vtype(func)

    def _add_type_check(self, func: gen.Block, anno_name: str):
        with func.b(f"if {self.VTYPE} is {anno_name}:") as b:
            b.l(f"{gen.Keyword.RET} {self.VNAME}")

    def _add_vtype(self, func: gen.Block):
        func.l(f"{self.VTYPE} = {self.VNAME}.__class__")

    def _get_default_factory(self, annotation: "AnnotationT"):
        factory: Union[Type, Callable[..., Any], None] = None
        args: Tuple = annotation.args if isinstance(annotation, Annotation) else tuple()
        if args:
            factory_anno = self.resolver.annotation(args[-1])
            if isinstance(factory_anno, ForwardDelayedAnnotation):
                return factory
            elif isinstance(factory_anno, DelayedAnnotation):
                use = factory_anno.type
                raw = use
            else:
                use = factory_anno.resolved_origin
                raw = factory_anno.un_resolved
            factory = use
            if issubclass(use, defaultdict):
                factory_nested = self._get_default_factory(factory_anno)

                def factory():
                    return defaultdict(factory_nested)

                factory.__qualname__ = f"factory({repr(raw)})"  # type: ignore

            if not checks.isbuiltinsubtype(use):  # type: ignore

                params: Mapping[str, inspect.Parameter] = cached_signature(
                    use
                ).parameters
                if not any(p.default is p.empty for p in params.values()):

                    def factory(*, __origin=use):
                        return __origin()

                    factory.__qualname__ = f"factory({repr(raw)})"  # type: ignore

        return factory

    def _build_text_des(self, context: BuildContext):
        func, annotation, anno_name = (
            context.func,
            context.annotation,
            context.anno_name,
        )
        origin = annotation.resolved_origin
        # Encode for bytes
        if issubclass(origin, bytes):
            with func.b(f"if isinstance({self.VNAME}, str):") as b:
                b.l(
                    f"{self.VNAME} = {anno_name}("
                    f"{self.VNAME}, encoding={DEFAULT_ENCODING!r})"
                )
        # Decode for str
        elif issubclass(origin, str):
            with func.b(f"if isinstance({self.VNAME}, (bytes, bytearray)):") as b:
                b.l(f"{self.VNAME} = {self.VNAME}.decode({DEFAULT_ENCODING!r})")
        func.l(f"{self.VNAME} = {anno_name}({self.VNAME})")

    def _build_builtin_des(self, context: BuildContext):
        func, annotation, anno_name = (
            context.func,
            context.annotation,
            context.anno_name,
        )
        origin = annotation.resolved_origin
        if issubclass(origin, (str, bytes)):
            self._build_text_des(context)
        elif checks.ismappingtype(origin):
            self._build_mapping_des(context)
        elif checks.iscollectiontype(origin):
            self._build_collection_des(context)
        # bool, int, float...
        else:
            func.l(f"{self.VNAME} = {anno_name}({self.VNAME})")

    def _build_pattern_des(self, context: BuildContext):
        func, anno_name = context.func, context.anno_name
        func.l(
            f"{self.VNAME} = {self.VNAME} "
            f"if issubclass({self.VTYPE}, {anno_name}) "
            f"else __re_compile({self.VNAME})",
            __re_compile=re.compile,
        )

    def _build_fromdict_des(self, func: gen.Block, anno_name: str):
        self._add_type_check(func, anno_name)
        func.l(f"{self.VNAME} = {anno_name}.from_dict({self.VNAME})")

    def _build_typeddict_des(self, context: BuildContext):
        func, annotation, namespace, anno_name = (
            context.func,
            context.annotation,
            context.namespace,
            context.anno_name,
        )
        total = getattr(context.annotation.resolved_origin, "__total__", True)
        with func.b(f"if issubclass({self.VTYPE}, Mapping):", Mapping=abc.Mapping) as b:
            fields_deser = {
                x: self.resolver._resolve_from_annotation(
                    y, namespace=namespace
                ).transmute
                for x, y in annotation.serde.fields.items()
            }
            x = "fields_in[x]"
            y = (
                f"fields_deser[x]({self.VNAME}[x])"
                if fields_deser
                else f"{self.VNAME}[x]"
            )
            line = f"{{{x}: {y} for x in fields_in.keys()"
            tail = "}" if total else f"& {self.VNAME}.keys()}}"
            b.l(
                f"{self.VNAME} = {anno_name}(**{line}{tail})", fields_deser=fields_deser
            )
        with func.b("else:") as b:
            b.l(
                f"{self.VNAME} = translate({self.VNAME}, {anno_name})",
                translate=self.resolver.translate,
            )

    def _build_typedtuple_des(self, context: BuildContext):
        func, annotation, anno_name = (
            context.func,
            context.annotation,
            context.anno_name,
        )
        with func.b(f"if issubclass({self.VTYPE}, Mapping):", Mapping=abc.Mapping) as b:
            if annotation.serde.fields:
                ctx = dataclasses.replace(context, func=b)
                self._build_typeddict_des(ctx)
            else:
                b.l(
                    f"{self.VNAME} = {anno_name}(**{self.VNAME})",
                )
        with func.b(
            f"elif isinstance({self.VNAME}, (list, set, frozenset, tuple)):"
        ) as b:
            if annotation.serde.fields:
                b.l(
                    f"{self.VNAME} = __bind({anno_name}, *{self.VNAME}).eval()",
                    __bind=self.resolver.bind,
                )
            else:
                b.l(
                    f"{self.VNAME} = {anno_name}(*{self.VNAME})",
                )
        with func.b("else:") as b:
            b.l(
                f"{self.VNAME} = translate({self.VNAME}, {anno_name})",
                translate=self.resolver.translate,
            )

    def _build_mapping_des(self, context: BuildContext):
        func, annotation, namespace, anno_name = (
            context.func,
            context.annotation,
            context.namespace,
            context.anno_name,
        )
        key_des, item_des = None, None
        args = annotation.args
        if args:
            args = cast(Tuple[Type, Type], args)
            key_type, item_type = args
            key_des = self.resolver.resolve(
                key_type, flags=annotation.serde.flags, namespace=namespace
            )
            item_des = self.resolver.resolve(
                item_type, flags=annotation.serde.flags, namespace=namespace
            )
        if issubclass(annotation.resolved_origin, defaultdict):
            factory = self._get_default_factory(annotation)
            func.namespace[anno_name] = functools.partial(defaultdict, factory)
        kd_name = f"{anno_name}_key_des"
        it_name = f"{anno_name}_item_des"
        iterate = f"iterate({self.VNAME})"
        line = f"{anno_name}({iterate})"
        if args or annotation.serde.fields_in:
            x, y = "x", "y"
            # If there are args & field mapping, get the correct field name
            # AND serialize the key.
            if args and annotation.serde.fields_in:
                x = f"{kd_name}(fields_in.get(x, x))"
            # If there is only a field mapping, get the correct name for the field.
            elif annotation.serde.fields_in:
                x = "fields_in.get(x, x)"
            # If there are only serializers, get the serialized value
            elif args:
                x = f"{kd_name}(x)"
                y = f"{it_name}(y)"
            line = f"{anno_name}({{{x}: {y} for x, y in {iterate}}})"
        # If we don't have nested annotations, we can short-circuit on valid inputs
        else:
            self._add_type_check(func, anno_name)
        # Write the lines.
        func.l(
            f"{self.VNAME} = {line}",
            level=None,
            **{
                kd_name: key_des,
                it_name: item_des,
                "Mapping": abc.Mapping,
                "iterate": self.resolver.iterate,
            },
        )

    def _build_tuple_des(self, context: BuildContext):
        func, annotation, namespace, anno_name = (
            context.func,
            context.annotation,
            context.namespace,
            context.anno_name,
        )
        if annotation.args and annotation.args[-1] is not ...:
            item_des = {
                ix: self.resolver.resolve(
                    t, flags=annotation.serde.flags, namespace=namespace
                )
                for ix, t in enumerate(annotation.args)
            }
            item_des_name = "item_des"
            iterate = f"iterate({self.VNAME}, values=True)"
            line = (
                f"{anno_name}"
                f"({item_des_name}[ix](v) for ix, v in enumerate({iterate})"
                f"if ix in {item_des_name})"
            )
            func.l(
                f"{self.VNAME} = {line}",
                level=None,
                **{
                    item_des_name: item_des,
                    "iterate": self.resolver.iterate,
                },
            )
        else:
            self._build_collection_des(context)

    def _build_collection_des(self, context: BuildContext):
        func, annotation, namespace, anno_name = (
            context.func,
            context.annotation,
            context.namespace,
            context.anno_name,
        )
        item_des = None
        it_name = f"{anno_name}_item_des"
        iterate = f"iterate({self.VNAME}, values=True)"
        line = f"{self.VNAME} = {anno_name}({iterate})"
        if annotation.args:
            item_type = annotation.args[0]
            item_des = self.resolver.resolve(
                item_type, flags=annotation.serde.flags, namespace=namespace
            )
            line = (
                f"{self.VNAME} = "
                f"{anno_name}({it_name}(x) for x in parent({iterate}))"
            )
        else:
            self._add_type_check(func, anno_name)
        func.l(
            line,
            level=None,
            **{
                it_name: item_des,
                "Collection": abc.Collection,
                "iterate": self.resolver.iterate,
            },
        )

    def _build_path_des(self, context: BuildContext):
        func, anno_name = context.func, context.anno_name
        self._add_type_check(func, anno_name)
        func.l(f"{self.VNAME} = {anno_name}({self.VNAME})")

    def _build_user_type_des(self, context: BuildContext):
        func, annotation, namespace, anno_name = (
            context.func,
            context.annotation,
            context.namespace,
            context.anno_name,
        )
        serde = annotation.serde
        resolved = annotation.resolved
        self._add_type_check(func, anno_name)
        # Main branch - we have a mapping for a user-defined class.
        # This is where the serde configuration comes in.
        # WINDY PATH AHEAD
        func.l("# Happy path - deserialize a mapping into the object.")
        with func.b(f"if issubclass({self.VTYPE}, Mapping):", Mapping=abc.Mapping) as b:
            # Universal line - transform input to known keys/values.
            # Specific values may change.
            def mainline(k, v):
                return f"{{{k}: {v} for x in fields_in.keys() & {self.VNAME}.keys()}}"

            # The "happy path" - e.g., no guesswork needed.
            def happypath(k, v, **ns):
                b.l(f"{self.VNAME} = {anno_name}(**{mainline(k, v)})", **ns)

            # Default X - translate given `x` to known input `x`
            x = "fields_in[x]"
            # No field name translation needs to happen.
            if {*serde.fields_in.keys()} == {*serde.fields_in.values()}:
                x = "x"

            # Default Y - get the given `y` with the given `x`
            y = f"{self.VNAME}[x]"
            # Get the intersection of known input fields and annotations.
            matched = {*serde.fields_in.values()} & serde.fields.keys()
            # Happy path! This is a `@typic.al` wrapped class.
            if self.resolver.known(resolved) or self.resolver.delayed(resolved):
                happypath(x, y)
            # Secondary happy path! We know how to deserialize already.
            else:
                fields_in = serde.fields_in
                fnamespace = namespace or resolved
                if serde.fields and len(matched) == len(serde.fields_in):
                    desers = {
                        f: self.resolver._resolve_from_annotation(
                            serde.fields[f], namespace=fnamespace
                        ).transmute
                        for f in matched
                    }
                else:
                    protocols = self.resolver.protocols(annotation.resolved_origin)
                    fields_in = {x: x for x in protocols}
                    desers = {f: p.transmute for f, p in protocols.items()}
                y = f"desers[{x}]({self.VNAME}[x])"
                happypath(x, y, desers=desers, fields_in=fields_in)

        # Secondary branch - we have some other input for a user-defined class
        func.l("# Unknown path, just try casting it directly.")
        with func.b(
            f"elif isbuiltinsubtype({self.VTYPE}):",
            isbuiltinsubtype=checks.isbuiltinsubtype,
        ) as b:
            b.l(f"{self.VNAME} = {anno_name}({self.VNAME})")
        # Final branch - user-defined class for another user-defined class
        func.l(
            "# Two user-defined types, "
            "try to translate the input into the desired output."
        )
        with func.b("else:") as b:
            b.l(
                f"{self.VNAME} = translate({self.VNAME}, {anno_name})",
                translate=self.resolver.translate,
            )

    def _build_literal_des(
        self, annotation: Annotation, func_name: str, namespace: Type = None
    ):
        args = annotation.args
        types: Set[Type] = {a.__class__ for a in args}
        t = types.pop() if len(types) == 1 else Union[tuple(types)]
        t_anno = cast(
            Annotation,
            self.resolver.annotation(
                t,  # type: ignore
                name=annotation.parameter.name,
                # parameter=annotation.parameter,
                is_optional=annotation.optional,
                is_strict=annotation.strict,
                flags=annotation.serde.flags,
                default=annotation.parameter.default,
            ),
        )
        return self._build_des(t_anno, func_name, namespace)

    def _build_union_des(self, context: BuildContext):
        func, annotation, namespace = (
            context.func,
            context.annotation,
            context.namespace,
        )
        # Get all types which we may coerce to.
        args = (*(a for a in annotation.args if a not in {None, Ellipsis, type(None)}),)
        if not args:
            return
        # Get all custom types, which may have discriminators
        targets = (*(a for a in args if not checks.isstdlibtype(a)),)
        # We can only build a tagged union deserializer if all args are valid
        if args != targets:
            return self._build_generic_union_des(context)

        # Try to collect the field which will be the discriminator.
        # First, get a mapping of Type -> Proto & Type -> Fields
        tagged = get_tag_for_types(targets)
        # Just bail out if we can't find a key.
        if not tagged:
            return self._build_generic_union_des(context)
        # If we got a key, re-map the protocols to the value for each type.
        deserializers = {
            value: self.resolver.resolve(t, namespace=namespace).transmute
            for value, t in tagged.types_by_values
        }
        # Finally, build the deserializer
        func.namespace.update(
            tag=tagged.tag,
            desers=deserializers,
            empty=_empty,
        )
        with func.b(f"if issubclass({self.VTYPE}, Mapping):", Mapping=abc.Mapping) as b:
            b.l(f"tag_value = {self.VNAME}.get(tag, empty)")
        with func.b("else:") as b:
            b.l(f"tag_value = getattr({self.VNAME}, tag, empty)")
        with func.b("if tag_value in desers:") as b:
            b.l(f"{self.VNAME} = desers[tag_value]({self.VNAME})")
        with func.b("else:") as b:
            b.l(
                "raise ValueError("
                'f"Value is missing field {tag!r} with one of '
                '{(*desers,)}: {val!r}"'
                ")"
            )

    def _build_generic_union_des(self, context: BuildContext):
        annotation, namespace, func = (
            context.annotation,
            context.namespace,
            context.func,
        )
        args = [a for a in annotation.args if a not in {None, Ellipsis, type(None)}]
        if args:
            ctx = {
                get_name(a)
                + "_des": self.resolver.resolve(a, namespace=namespace).transmute
                for a in args
            }
            names = (*(n.rpartition("_")[0] for n in ctx),)
            for name, call in ctx.items():
                with func.b("try:") as b:
                    b.l(f"return {name}({self.VNAME})")
                with func.b("except (TypeError, ValueError, KeyError):") as b:
                    b.l("pass")
            func.namespace.update(ctx)
            func.l(
                "raise ValueError("
                'f"Value could not be deserialized into one of {names}: {val!r}"'
                ")",
                names=names,
            )
            return False

    def _build_des(  # noqa: C901
        self,
        annotation: Annotation[Type[ObjectT]],
        func_name: str,
        namespace: Type = None,
    ) -> DeserializerT[ObjectT]:
        args = annotation.args
        # Get the "origin" of the annotation.
        # For natives and their typing.* equivs, this will be a builtin type.
        # For SpecialForms (Union, mainly) this will be the un-subscripted type.
        # For custom types or classes, this will be the same as the annotation.
        origin = annotation.resolved_origin
        anno_name = get_unique_name(origin)
        ns = {
            anno_name: origin,
            "parent": getattr(origin, "__parent__", origin),
            "issubclass": cached_issubclass,
            **annotation.serde.asdict(),
        }
        if checks.isliteral(origin):
            return self._build_literal_des(annotation, func_name, namespace)
        with gen.Block(ns) as main:
            with main.f(func_name, main.param(f"{self.VNAME}")) as func:
                needs_return = None
                context = BuildContext(annotation, ns, anno_name, func, namespace)
                if origin not in self.UNRESOLVABLE:
                    # Set our top-level sanity checks.
                    self._set_checks(func, anno_name, annotation)
                    # Move through our queue.
                    for check, handler in self._HANDLERS.items():
                        # If this is a valid type for this handler,
                        #   write the deserializer.
                        if check(origin, args):
                            needs_return = handler(self, context)
                            break
                # If the deserializer doesn't contain a return statement, add one.
                if needs_return is not False:
                    func.l(f"{gen.Keyword.RET} {self.VNAME}")
        deserializer = main.compile(ns=ns, name=func_name)
        return deserializer

    # Order is IMPORTANT! This is a FIFO queue.
    _HANDLERS: Mapping[HandlerCheckT, BuildHandlerT] = {
        # Special handler for Unions...
        lambda origin, args: checks.isuniontype(origin): _build_union_des,
        # Non-intersecting types (order doesn't matter here.
        lambda origin, args: checks.isdatetype(origin): _build_date_des,
        lambda origin, args: checks.istimetype(origin): _build_time_des,
        lambda origin, args: checks.istimedeltatype(origin): _build_timedelta_des,
        lambda origin, args: checks.isuuidtype(origin): _build_uuid_des,
        lambda origin, args: origin in {Pattern, re.Pattern}: _build_pattern_des,
        lambda origin, args: issubclass(origin, pathlib.Path): _build_path_des,
        # MUST come before subtype check.
        lambda origin, args: (
            not args and checks.isbuiltintype(origin)
        ): _build_builtin_des,
        # Psuedo-structured containers, should check before generics.
        lambda origin, args: checks.istypeddict(origin): _build_typeddict_des,
        lambda origin, args: checks.istypedtuple(origin): _build_typedtuple_des,
        lambda origin, args: checks.isnamedtuple(origin): _build_typedtuple_des,
        lambda origin, args: (
            not args and checks.isbuiltinsubtype(origin)
        ): _build_builtin_des,
        # A mapping is a collection so must come before that check.
        lambda origin, args: checks.ismappingtype(origin): _build_mapping_des,
        # A tuple is a collection so must come before that check.
        lambda origin, args: checks.istupletype(origin): _build_tuple_des,
        # Generic collection handler
        lambda origin, args: checks.iscollectiontype(origin): _build_collection_des,
        # Catch-all for custom user types (user-defined classes).
        lambda origin, args: True: _build_user_type_des,
    }

    def _check_varargs(
        self,
        anno: Annotation[Type[ObjectT]],
        des: DeserializerT[ObjectT],
        validator: const.ValidatorT[ObjectT],
    ) -> Tuple[DeserializerT[ObjectT], const.ValidatorT[ObjectT]]:
        if anno.parameter.kind == VAR_POSITIONAL:
            __des = des
            __validator = validator

            def des(__val, *, __des=__des):  # type: ignore
                return (*(__des(x) for x in __val),)

            def validator(value, *, field: str = None, __validator=__validator):  # type: ignore
                return (*(__validator(x, field=field) for x in value),)

        elif anno.parameter.kind == VAR_KEYWORD:
            __des = des
            __validator = validator

            def des(__val, *, __des=__des):  # type: ignore
                return {x: __des(y) for x, y in __val.items()}

            def validator(value, *, field: str = None, __validator=__validator):  # type: ignore
                return {x: __validator(y, field=field) for x, y in value.items()}

        return des, validator

    def _finalize_validator(
        self,
        constr: Optional[const.ConstraintsProtocolT[ObjectT]],
    ) -> const.ValidatorT[ObjectT]:
        def validate(value, *, field: str = None):
            return value

        if constr:
            validate = constr.validate  # noqa: F811

        return validate  # type: ignore

    def _finalize_deserializer(
        self,
        anno: Annotation[Type[ObjectT]],
        des: DeserializerT[ObjectT],
        constr: Optional[const.ConstraintsProtocolT[ObjectT]],
    ) -> Tuple[DeserializerT[ObjectT], const.ValidatorT[ObjectT]]:
        # Determine how to run in "strict-mode"
        validator = self._finalize_validator(constr)
        # Handle *args and **kwargs
        des, validator = self._check_varargs(anno, des, validator)
        # If we have type constraints, override the deserializer for strict annotations.
        if isinstance(constr, (const.TypeConstraints, const.LiteralConstraints)):
            if anno.strict:
                des = validator  # type: ignore
            elif isinstance(constr, const.LiteralConstraints):
                __d = des

                def des(val: Any, *, __d=__d, __v=validator) -> ObjectT:
                    return __v(__d(val))

        # Otherwise
        else:
            # In strict mode, we validate & coerce if there are constraints
            if anno.strict and constr and constr.coerce:
                __d = des

                def des(val: Any, *, __d=__d, __v=validator) -> ObjectT:  # type: ignore
                    return __d(__v(val))

            elif anno.strict:
                des = validator  # type: ignore

        return des, validator

    def factory(
        self,
        annotation: Annotation[Type[ObjectT]],
        constr: Optional[const.ConstraintsProtocolT[ObjectT]] = None,
        namespace: Type = None,
    ) -> Tuple[DeserializerT[ObjectT], const.ValidatorT[ObjectT]]:
        annotation.serde = annotation.serde or SerdeConfig()
        key = self._get_name(annotation, constr)
        if key in self.__DES_CACHE:
            return self.__DES_CACHE[key]
        deserializer: Optional[DeserializerT] = None
        for check, des in self.__USER_DESS:
            if check(annotation.resolved):
                deserializer = des
                break
        if not deserializer:
            deserializer = self._build_des(annotation, key, namespace)

        deserializer, validator = self._finalize_deserializer(
            annotation, deserializer, constr
        )
        self.__DES_CACHE[key] = (deserializer, validator)

        return deserializer, validator


@slotted(dict=False, weakref=True)
@dataclasses.dataclass
class BuildContext:
    annotation: Annotation
    ns: Mapping[str, Any]
    anno_name: str
    func: gen.Block
    namespace: Optional[Type] = None


HandlerCheckT = Callable[[Type[ObjectT], Tuple[Any, ...]], TypeGuard[Type[ObjectT]]]
BuildHandlerT = Callable[[DesFactory, BuildContext], Union[NoReturn, Literal[False]]]



from __future__ import annotations

import pathlib


__all__ = ("FilePathError", "FilePath", "DirectoryPathError", "DirectoryPath")


PathType = pathlib.Path().__class__  # type: ignore


class FilePathError(ValueError):
    """Generic error raised when the given value is not a File-Path."""

    pass


class FilePath(PathType):  # type: ignore
    """A path object pointing to a file.

    See Also
    --------
    :py:class:`pathlib.Path`
    """

    def __init__(self, *pathsegments: str):
        super().__init__()
        if not self.is_file():
            raise FilePathError(f"{self} is not a valid file-path") from None


class DirectoryPathError(ValueError):
    """Generic error raised when the given value is not a Directory-Path."""

    pass


class DirectoryPath(PathType):  # type: ignore
    """A path object pointing to a directory.

    See Also
    --------
    :py:class:`pathlib.Path`
    """

    def __init__(self, *pathsegments: str):
        super().__init__()
        if not self.is_dir():
            raise DirectoryPathError(f"{self} is not a valid directory-path") from None



from __future__ import annotations

import dataclasses
import re
from typing import ClassVar, Pattern, Match, Optional
from urllib.parse import quote

from typic.util import cached_property, slotted
from .url import (
    NetworkAddress,
    PRIVATE_HOSTS,
    INTERNAL_HOSTS,
    INTERNAL_IP_PATTERN,
    NetworkAddressValueError,
)

__all__ = ("Email", "EmailAddrInfo", "EMAIL_PATTERN", "EmailValueError")


class EmailValueError(NetworkAddressValueError):
    """A generic error for when we've received an invalid value for an email."""

    pass


# http://emailregex.com/
# https://help.returnpath.com/hc/en-us/articles/220560587-What-are-the-rules-for-email-address-syntax-
# expanding a bit for more specific detection
# also violating DRY :'( by copying the host regex from URL,
# but hey, this means we're more compliant with RFC 5322,
# AND our regex is readable, and readability counts!
EMAIL_PATTERN = re.compile(
    r"""
    (
        ^
        ((?P<name>([A-Z]+\s?)+)\s<)?
        # user
        (?P<username>([A-Z0-9]([!#$%&'*=?^`{|_.+-])?)*[A-Z0-9]+)
        @
        # host
        (?P<host>(?:
        # Domain
            (?P<domain>
                (?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+
                (?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)
            )
            # Localhost
            |(?P<localhost>localhost)
            |(?P<dotless>(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.?))
            # IPV4
            |(?P<ipv4>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})
            # IPV6
            |(?P<ipv6>\[[A-F0-9]*:[A-F0-9:]+\])
        ))
        (>)?
        $
    )
    """,
    re.I | re.VERBOSE,
)


@slotted(dict=True)
@dataclasses.dataclass(frozen=True)
class EmailAddrInfo:
    """Detailed information about an email address.

    Can be called directly, generated by casting a :py:class:`str` as :py:class:`Email`,
    or created with :py:meth:`EmailAddrInfo.from_str`

    Notes
    -----
    An email address is also, technically, a network address. However, it has some more
    constraints and we're interested in only a specific amount of information.
    """

    PATTERN: ClassVar[Pattern] = EMAIL_PATTERN
    """The raw match from paring the provided string."""
    name: str
    """The proper name associated to the email."""
    username: str
    """The username portion of the email."""
    host: str
    """The host address where the email server is located."""
    is_ip: bool = False

    @classmethod
    def from_str(cls, value) -> "EmailAddrInfo":
        """Parse & validate a string, generate an instance of :py:class:`EmailAddrInfo`."""
        match: Optional[Match] = cls.PATTERN.match(value)
        if not match or not value:
            err_msg = f"<{value!r}> is not a valid email address."
            raise EmailValueError(err_msg) from None
        return cls(
            name=(match["name"] or "").rstrip(),
            username=match["username"],
            host=match["host"],
            is_ip=bool(match["ipv4"] or match["ipv6"]),
        )

    @cached_property
    def address(self) -> str:
        """The fully-qualified email address.

        If this instance was generated from a string, it will match.
        """
        name = f"{self.name} <" if self.name else ""
        address = f"{self.username}@{self.host}"
        return f"{name}{address}>" if name else address

    @cached_property
    def address_encoded(self) -> str:
        """The fully-qualified email address, encoded."""
        name = f"{self.name} <" if self.name else ""
        address = quote(f"{self.username}@{self.host}")
        return f"{name}{address}>" if name else address

    @cached_property
    def is_named(self) -> bool:
        """Whether or not this email is 'named' (or 'pretty')

        i.e.: `<Foo foo@bar.com>`
        """
        return bool(self.name)

    @cached_property
    def is_private(self) -> bool:
        """Whether or not the host in the email is a private IP, i.e., 'localhost'."""
        return self.host in PRIVATE_HOSTS

    @cached_property
    def is_internal(self) -> bool:
        """Whether or not the host in the email is an internal address.

        Internal DNS/IP addresses aren't necessarily private, hence the differentiation.
        """
        return bool(
            self.host in INTERNAL_HOSTS
            or (self.is_ip and INTERNAL_IP_PATTERN.match(self.host))
        )


# Deepcopy is broken for frozen dataclasses with slots.
# https://github.com/python/cpython/pull/17254
# EmailAddrInfo.__slots__ = tuple(_.name for _ in dataclasses.fields(EmailAddrInfo))


class Email(NetworkAddress):
    """An immutable email address. Supports 'pretty' and 'raw', i.e.:

        `Foo Bar <foo.bar@foobar.net>`
        `foo.bar@foobar.net`

    Detailed information about the email string can be found up via :py:attr:`Email.info`.

    Examples
    --------
    >>> import typic
    >>> email = typic.Email("Foo Bar <foo.bar@foobar.net>")
    >>> print(email)
    Foo Bar <foo.bar@foobar.net>
    >>> email.info.host
    'foobar.net'
    >>> email.info.is_named
    True
    >>> import json
    >>> json.dumps([email])
    '["Foo Bar <foo.bar@foobar.net>"]'

    See Also
    --------
    :py:class:`EmailAddrInfo`

    Notes
    -----
    This object inherits from :py:class:`str` and so is natively JSON-serializable.
    """

    @cached_property
    def info(self) -> EmailAddrInfo:  # type: ignore
        """Get detailed information about your email string.

        See Also
        --------
        :py:class:`EmailAddrInfo`
        """

        return EmailAddrInfo.from_str(self)



from __future__ import annotations

import dataclasses
import datetime
import decimal
import enum
import inspect
import ipaddress
import pathlib
import re
import uuid
from collections import abc
from collections.abc import (
    Mapping as Mapping_abc,
    Collection as Collection_abc,
    Iterable as Iterable_abc,
)
from operator import methodcaller, attrgetter
from types import MappingProxyType
from typing import (
    Type,
    Callable,
    Collection,
    Union,
    Mapping,
    Any,
    ClassVar,
    cast,
    TYPE_CHECKING,
    Iterable,
    MutableMapping,
    Dict,
    Optional,
    TypeVar,
)

from typic import util, checks, gen, types
from typic.common import DEFAULT_ENCODING
from typic.compat import Literal
from .common import (
    SerializerT,
    SerdeConfig,
    Annotation,
    ForwardDelayedAnnotation,
    DelayedAnnotation,
)

if TYPE_CHECKING:  # pragma: nocover
    from .resolver import Resolver


class SerializationValueError(ValueError):
    ...


_decode = methodcaller("decode", DEFAULT_ENCODING)
_pattern = attrgetter("pattern")
_T = TypeVar("_T")


class SerFactory:
    """A factory for generating high-performance serializers.
    Notes
    -----
    Should not be used directly.
    """

    _DEFINED: Mapping[Type, SerializerT] = {
        ipaddress.IPv4Address: str,
        ipaddress.IPv4Network: str,
        ipaddress.IPv6Address: str,
        ipaddress.IPv6Interface: str,
        ipaddress.IPv6Network: str,
        re.Pattern: _pattern,  # type: ignore
        pathlib.Path: str,
        types.AbsoluteURL: str,
        types.DSN: str,
        types.DirectoryPath: str,
        types.Email: str,
        types.FilePath: str,
        types.HostName: str,
        types.NetworkAddress: str,
        types.RelativeURL: str,
        types.SecretBytes: cast(SerializerT, lambda o: _decode(o.secret)),
        types.SecretStr: cast(SerializerT, attrgetter("secret")),
        types.URL: str,
        uuid.UUID: str,
        decimal.Decimal: float,
        bytes: cast(SerializerT, _decode),
        bytearray: cast(SerializerT, _decode),
        datetime.date: cast(SerializerT, util.isoformat),
        datetime.datetime: cast(SerializerT, util.isoformat),
        datetime.time: cast(SerializerT, util.isoformat),
        datetime.timedelta: cast(SerializerT, util.isoformat),
    }

    _LISTITER = (
        list,
        tuple,
        set,
        frozenset,
        Collection,
        Collection_abc,
        Iterable,
        Iterable_abc,
    )
    _DICTITER = (dict, Mapping, Mapping_abc, MappingProxyType, types.FrozenDict)
    _PRIMITIVES = (str, int, bool, float, type(None), type(...))
    _DYNAMIC = frozenset(
        {Union, Any, inspect.Parameter.empty, dataclasses.MISSING, ClassVar, Literal}
    )
    _FNAME = "fname"

    def __init__(self, resolver: Resolver):
        self.resolver = resolver
        self._serializer_cache: MutableMapping[str, SerializerT] = {}

    @staticmethod
    def _get_name(annotation: Annotation) -> str:
        return util.get_defname("serializer", annotation)

    def _check_add_null_check(self, func: gen.Function, annotation: Annotation):
        if annotation.optional:
            with func.b(f"if o in {self.resolver.OPTIONALS}:") as b:
                b.l(f"{gen.Keyword.RET}")

    def _add_type_check(self, func: gen.Function, annotation: Annotation):
        resolved_name = util.get_name(annotation.resolved)
        func.l(f"{self._FNAME} = name or {resolved_name!r}")
        line = "if not tcheck(o.__class__, t):"
        check: Callable[..., bool] = util.cached_issubclass
        t = annotation.generic
        if checks.isbuiltinsubtype(annotation.generic):
            line = "if not tcheck(o, t):"
            check = isinstance  # type: ignore
        if checks.istypeddict(t):
            t = dict
        with func.b(line, tcheck=check) as b:
            msg = (
                f"{{{self._FNAME}}}: type {{inst_tname!r}} "
                f"is not a subtype of type "
                f"{util.get_qualname(annotation.generic)!r}. "
                f"Perhaps this annotation should be "
                f"Union[{{inst_tname}}, {util.get_qualname(annotation.generic)}]?"
            )
            b.l("inst_tname = qualname(o.__class__)")
            b.l(
                f"raise err(f{msg!r})",
                err=SerializationValueError,
                qualname=util.get_qualname,
                t=t,
            )

    def _build_list_serializer(
        self,
        func: gen.Function,
        annotation: Annotation,
    ):
        # Check for value types
        line = "[*o]"
        ns: Dict[str, Any] = {}
        self._check_add_null_check(func, annotation)
        self._add_type_check(func, annotation)
        if annotation.args:
            arg_a: Annotation = self.resolver.annotation(
                annotation.args[0], flags=annotation.serde.flags
            )
            arg_ser = self.factory(arg_a)
            arg_ser_name = "arg_ser"
            ns[arg_ser_name] = arg_ser
            func.l(f"gen = ({arg_ser_name}(v) for v in o)")
            line = "gen if lazy else [*gen]"

        func.l(f"{gen.Keyword.RET} {line}", level=None, **ns)

    def _build_dict_serializer(self, func: gen.Function, annotation: Annotation):
        # Check for args
        kser_: SerializerT
        vser_: SerializerT
        kser_, vser_ = cast(SerializerT, self.resolver.primitive), cast(
            SerializerT, self.resolver.primitive
        )
        args = util.get_args(annotation.resolved)
        if args:
            kt, vt = args
            ktr: Annotation = self.resolver.annotation(kt, flags=annotation.serde.flags)
            vtr: Annotation = self.resolver.annotation(vt, flags=annotation.serde.flags)
            kser_, vser_ = (self.factory(ktr), self.factory(vtr))
        # Add sanity checks.
        self._check_add_null_check(func, annotation)
        self._add_type_check(func, annotation)
        ns: Dict[str, Any] = {
            "kser": kser_,
            "vser": vser_,
        }
        ksercall = "kser(k)"
        if annotation.serde.fields_out:
            ns["fields_out"] = annotation.serde.fields_out
            ksercall = "kser(fields_out.get(k, k))"
        if annotation.serde.flags.case:
            ns["case"] = annotation.serde.flags.case.transformer
            ksercall = f"case({ksercall})"
        gencall = f"({ksercall}, vser(v)) for k, v in o.items()"
        if annotation.serde.flags.omit:
            ns["omit"] = annotation.serde.flags.omit
            gencall = f"{gencall} if v not in omit"
        func.l(
            f"gen = ({gencall})",
            **ns,
        )
        func.l(f"{gen.Keyword.RET} gen if lazy else {{k: v for k, v in gen}}")

    def _build_class_serializer(
        self,
        func: gen.Function,
        annotation: Annotation,
    ):
        # Get the field serializers
        fields_ser = {x: self.factory(y) for x, y in annotation.serde.fields.items()}
        iterator = self.resolver.translator.iterator(
            annotation.resolved,
            relaxed=True,
            # We want to proactively exclude defined fields from this iterator.
            exclude=(*annotation.serde.flags.exclude,),
        )
        self._check_add_null_check(func, annotation)
        self._add_type_check(func, annotation)
        ns: Dict[str, Any] = {
            "fields_ser": fields_ser,
            "fields_out": annotation.serde.fields_out.keys(),
            "iterator": iterator,
        }
        # Determine how to dump the field.
        f = "f"
        # Get the mapping of attr->out
        # If we have transforms, make sure we use them on dump.
        transforms = {f: t for f, t in annotation.serde.fields_out.items() if f != t}
        if transforms:
            ns["transforms"] = transforms
            f = "transforms.get(f, f)"
        # Define the generator expression.
        gencall = f"({f}, fields_ser[f](v)) for f, v in iterator(o) if f in fields_out"
        if annotation.serde.flags.omit:
            ns["omit"] = annotation.serde.flags.omit
            gencall = f"{gencall} and v not in omit"

        func.l(f"gen = ({gencall})", **ns)
        func.l(f"{gen.Keyword.RET} gen if lazy else {{f: v for f, v in gen}}")

    def _compile_enum_serializer(self, annotation: Annotation) -> SerializerT:
        origin: Type[enum.Enum] = cast(Type[enum.Enum], annotation.resolved_origin)
        ts = {type(x.value) for x in origin}
        # If we can predict a single type the return the serializer for that
        if len(ts) == 1:
            t = ts.pop()
            va = self.resolver.annotation(t, flags=annotation.serde.flags)
            vser = self.factory(va)

            def serializer(
                o: enum.Enum, *, lazy: bool = False, name: util.ReprT = None, _vser=vser
            ):
                return _vser(o.value, lazy=lazy, name=name)

            return cast(SerializerT, serializer)
        # Else default to lazy serialization
        return cast(SerializerT, self.resolver.primitive)

    def _compile_defined_serializer(
        self,
        annotation: Annotation[Type[_T]],
        ser: SerializerT[_T],
    ) -> SerializerT[_T]:
        func_name = self._get_name(annotation)
        ser_name = "ser"
        ns = {ser_name: ser}
        with gen.Block(ns) as main:
            with self._define(main, func_name) as func:
                self._check_add_null_check(func, annotation)
                self._add_type_check(func, annotation)
                line = f"{ser_name}(o)"
                if annotation.origin in (type(o) for o in self.resolver.OPTIONALS):
                    line = "None"
                func.l(f"{gen.Keyword.RET} {line}")

        serializer: SerializerT = main.compile(name=func_name, ns=ns)
        return serializer

    def _compile_defined_subclass_serializer(
        self, origin: Type, annotation: Annotation
    ):
        for t, s in self._DEFINED.items():
            if issubclass(origin, t):
                return self._compile_defined_serializer(annotation, s)
        # pragma: nocover

    def _compile_primitive_subclass_serializer(
        self, origin: Type, annotation: Annotation
    ):
        for t in self._PRIMITIVES:
            if issubclass(origin, t):
                return self._compile_defined_serializer(annotation, t)
        # pragma: nocover

    def _compile_serializer(self, annotation: Annotation[Type[_T]]) -> SerializerT[_T]:
        # Check for an optional and extract the type if possible.
        func_name = self._get_name(annotation)
        # We've been here before...
        if func_name in self._serializer_cache:
            return self._serializer_cache[func_name]

        serializer: SerializerT
        origin = annotation.resolved_origin
        # Lazy shortcut for messy paths (Union, Any, ...)
        if origin in self._DYNAMIC or not annotation.static:
            serializer = cast(SerializerT, self.resolver.primitive)
        # Routines (functions or methods) can't be serialized...
        elif issubclass(origin, abc.Callable) or inspect.isroutine(origin):  # type: ignore
            name = util.get_qualname(origin)
            with gen.Block() as main:
                with self._define(main, func_name) as func:
                    func.l(
                        f'raise TypeError("Routines are not serializable. ({name!r}).")'
                    )

            serializer = main.compile(name=func_name)
            self._serializer_cache[func_name] = serializer
        # Enums are special
        elif checks.isenumtype(annotation.resolved):
            serializer = self._compile_enum_serializer(annotation)
        # Primitives don't require further processing.
        # Just check for nullable and the correct type.
        elif origin in self._PRIMITIVES:
            ns: dict = {}
            with gen.Block(ns) as main:
                with self._define(main, func_name) as func:
                    self._check_add_null_check(func, annotation)
                    self._add_type_check(func, annotation)
                    line = "o"
                    if annotation.origin in (type(o) for o in self.resolver.OPTIONALS):
                        line = "None"
                    func.l(f"{gen.Keyword.RET} {line}")

            serializer = main.compile(name=func_name, ns=ns)
            self._serializer_cache[func_name] = serializer

        # Defined cases are pre-compiled, but we have to check for optionals.
        elif origin in self._DEFINED:
            serializer = self._compile_defined_serializer(
                annotation, self._DEFINED[origin]
            )
        elif issubclass(origin, (*self._DEFINED,)):
            serializer = self._compile_defined_subclass_serializer(origin, annotation)
        elif issubclass(origin, self._PRIMITIVES):
            serializer = self._compile_primitive_subclass_serializer(origin, annotation)
        else:
            # Build the function namespace
            anno_name = f"{func_name}_anno"
            ns = {anno_name: origin, **annotation.serde.asdict()}
            with gen.Block(ns) as main:
                with self._define(main, func_name) as func:
                    # Mapping types need special nested processing as well
                    istypeddict = checks.istypeddict(origin)
                    istypedtuple = checks.istypedtuple(origin)
                    if not istypeddict and issubclass(origin, self._DICTITER):
                        self._build_dict_serializer(func, annotation)
                    # Array types need nested processing.
                    elif (
                        not istypedtuple
                        and not istypeddict
                        and issubclass(origin, self._LISTITER)
                    ):
                        self._build_list_serializer(func, annotation)
                    # Build a serializer for a structured class.
                    else:
                        self._build_class_serializer(func, annotation)
            serializer = main.compile(name=func_name, ns=ns)
            self._serializer_cache[func_name] = serializer
        return serializer

    @staticmethod
    def _define(main: gen.Block, name: str) -> gen.Function:
        return main.func(
            name,
            main.param("o"),
            main.param("lazy", default=False, kind=gen.ParameterKind.KEYWORD_ONLY),
            main.param("name", default=None, kind=gen.ParameterKind.KEYWORD_ONLY),
        )

    def factory(self, annotation: Annotation[Type[_T]]) -> SerializerT[_T]:
        if isinstance(annotation, (DelayedAnnotation, ForwardDelayedAnnotation)):
            return cast(SerializerT, DelayedSerializer(annotation, self))
        annotation.serde = annotation.serde or SerdeConfig()
        return self._compile_serializer(annotation)


class DelayedSerializer:
    __slots__ = "anno", "factory", "_serializer", "__name__"

    def __init__(
        self,
        anno: Union[DelayedAnnotation, ForwardDelayedAnnotation],
        factory: SerFactory,
    ):
        self.anno = anno
        self.factory = factory
        self._serializer: Optional[SerializerT] = None
        self.__name__ = anno.name

    def __call__(self, *args, **kwargs):
        if self._serializer is None:
            self._serializer = self.factory.factory(self.anno.resolved.annotation)
        return self._serializer(*args, **kwargs)



from __future__ import annotations as a

import collections
import dataclasses
import functools
import inspect
from types import FunctionType, MethodType
from typing import (
    Union,
    Callable,
    Type,
    Tuple,
    Optional,
    Mapping,
    TypeVar,
    cast,
    Iterable,
    Any,
    Dict,
    Set,
    List,
)

import typic.constraints as c
from typic.checks import issubclass, isfrozendataclass
from typic.compat import lru_cache
from typic.env import Environ, EnvironmentTypeError, EnvironmentValueError
from typic.serde.binder import BoundArguments
from typic.serde.common import (
    Annotation,
    SerdeFlags,
    SerializerT,
    SerdeProtocol,
    SerdeProtocolsT,
    DeserializerT,
    TranslatorT,
    FieldIteratorT,
)
from typic.common import (
    ORIG_SETTER_NAME,
    SCHEMA_NAME,
    SERDE_FLAGS_ATTR,
    Case,
    ReadOnly,
    WriteOnly,
    SERDE_ATTR,
    TYPIC_ANNOS_NAME,
)
from typic.serde.resolver import resolver
from typic.serde.ser import SerializationValueError
from typic.strict import (
    is_strict_mode,
    strict_mode,
    Strict,
    StrictStrT,
    STRICT_MODE,
    StrictModeT,
)
from typic.ext.schema import SchemaFieldT, builder as schema_builder, ObjectSchemaField
from typic.util import origin, cached_type_hints
from typic.types import FrozenDict, freeze

__all__ = (
    "Annotation",
    "annotations",
    "bind",
    "BoundArguments",
    "Case",
    "coerce",
    "constrained",
    "decode",
    "encode",
    "environ",
    "EnvironmentTypeError",
    "EnvironmentValueError",
    "flags",
    "is_strict_mode",
    "iterate",
    "tojson",
    "primitive",
    "protocol",
    "protocols",
    "ReadOnly",
    "register",
    "resolve",
    "resolver",
    "settings",
    "schema",
    "schemas",
    "SchemaReturnT",
    "SerdeFlags",
    "SerdeProtocol",
    "SerializationValueError",
    "Strict",
    "strict_mode",
    "StrictStrT",
    "transmute",
    "translate",
    "typed",
    "validate",
    "wrap",
    "wrap_cls",
    "WriteOnly",
)

ObjectT = TypeVar("ObjectT")
SchemaGenT = Callable[[Type[ObjectT]], SchemaFieldT]
_TO_RESOLVE: Set[Union[Type["WrappedObjectT"], Callable]] = set()


transmute = resolver.transmute
translate = resolver.translate
validate = resolver.validate
bind = resolver.bind
register = resolver.des.register
primitive = resolver.primitive
schemas = schema_builder.all
protocols = resolver.protocols
protocol = resolver.resolve
tojson = resolver.tojson
iterate = resolver.iterate
flags = SerdeFlags
encode = resolver.encode
decode = resolver.decode

# TBDeprecated
coerce = resolver.coerce_value
annotations = resolver.protocols


_T = TypeVar("_T")


class TypicObjectT:
    __serde__: SerdeProtocol
    __serde_flags__: SerdeFlags
    __serde_protocols__: SerdeProtocolsT
    __setattr_original__: Callable[["WrappedObjectT", str, Any], None]
    __typic_resolved__: bool
    __iter__: FieldIteratorT
    schema: SchemaGenT
    primitive: SerializerT
    transmute: DeserializerT
    translate: TranslatorT
    validate: "c.ValidatorT"
    tojson: Callable[..., str]
    iterate: FieldIteratorT


WrappedObjectT = Union[TypicObjectT, ObjectT]


def wrap(
    func: Callable[..., _T], *, delay: bool = False, strict: StrictModeT = STRICT_MODE
) -> Callable[..., _T]:
    """Wrap a callable to automatically enforce type-coercion.

    Parameters
    ----------
    func
        The callable for which you wish to ensure type-safety
    delay
        Delay annotation resolution until the first call
    strict
        Turn on "validator mode": e.g. validate incoming data rather than coerce.

    See Also
    --------
    :py:func:`inspect.signature`
    :py:meth:`inspect.Signature.bind`
    """
    if not delay:
        protocols(func)
    else:
        _TO_RESOLVE.add(func)

    @functools.wraps(func)
    def func_wrapper(*args, **kwargs) -> _T:
        bound = bind(func, *args, strict=strict, **kwargs)
        return bound.eval()

    return func_wrapper


def _bind_proto(cls, proto: SerdeProtocol):
    for n, attr in (
        (SERDE_ATTR, proto),
        ("primitive", proto.primitive),
        ("tojson", proto.tojson),
        ("transmute", staticmethod(proto.transmute)),
        ("validate", staticmethod(proto.validate)),
        ("translate", proto.translate),
        ("encode", proto.encode),
        ("decode", staticmethod(proto.decode)),
        ("iterate", proto.iterate),
        ("__iter__", proto.iterate),
    ):
        setattr(cls, n, attr)


@lru_cache(maxsize=None)
def _resolve_class(
    cls: Type[ObjectT],
    *,
    strict: StrictModeT = STRICT_MODE,
    always: bool = True,
    jsonschema: bool = True,
    serde: SerdeFlags = None,
) -> Type[WrappedObjectT]:
    # Build the namespace for the new class
    strict = cast(bool, strict)
    protos = protocols(cls, strict=strict)
    if hasattr(cls, SERDE_FLAGS_ATTR):
        pserde: SerdeFlags = getattr(cls, SERDE_FLAGS_ATTR)
        serde = pserde.merge(serde) if serde else pserde
    serde = serde or SerdeFlags()
    ns: Dict[str, Any] = {
        SERDE_FLAGS_ATTR: serde,
        TYPIC_ANNOS_NAME: protos,
    }
    if jsonschema:
        ns["schema"] = classmethod(schema)
        schema_builder.attach(cls)

    # Frozen dataclasses don't use the native setattr
    # So we wrap the init. This should be fine,
    # just slower :(
    if isfrozendataclass(cls):
        ns["__init__"] = wrap(cls.__init__, strict=strict)
    # The faster way - create a new setattr that applies the protocol for a given attr
    else:
        trans = freeze({x: y.transmute for x, y in protos.items()})

        def setattr_typed(setter):
            @functools.wraps(setter)
            def __setattr_typed__(self, name, item, *, __trans=trans, __setter=setter):
                __setter(
                    self,
                    name,
                    __trans[name](item) if name in __trans else item,
                )

            return __setattr_typed__

        ns.update(
            **{
                ORIG_SETTER_NAME: _get_setter(cls),
                "__setattr__": setattr_typed(cls.__setattr__),
            }
        )

    for name, attr in ns.items():
        setattr(cls, name, attr)
    # Get the protocol
    proto: SerdeProtocol = resolver.resolve(cls, is_strict=strict)
    # Bind it to the new class
    _bind_proto(cls, proto)
    # Track resolution state.
    setattr(cls, "__typic_resolved__", True)
    return cls


def _delay_resolve_class(
    cls_: Type[ObjectT],
    *,
    strict: StrictModeT = STRICT_MODE,
    jsonschema: bool = True,
    serde: SerdeFlags = None,
):
    # This class ain't resolved yet.
    setattr(cls_, "__typic_resolved__", False)

    # Create a classmethod for delayed resolution.
    def _resolve(cls):
        if not cls.__typic_resolved__:
            _resolve_class(cls, strict=strict, jsonschema=jsonschema, serde=serde)
            if cls in _TO_RESOLVE:
                _TO_RESOLVE.remove(cls)
        cls.__typic_resolved__ = True

    setattr(cls_, "resolve", classmethod(_resolve))

    # Allow users to delay until first init
    # This adds some extra cost on init that
    # It's not too much but something to consider
    def init(_init):
        @functools.wraps(_init)
        def __delayed_init(*args, **kwargs):
            cls_.resolve()
            _init(*args, **kwargs)

        return __delayed_init

    setattr(cls_, "__init__", init(cls_.__init__))

    # Add the cls to the stack of those to resolve
    _TO_RESOLVE.add(cls_)


def wrap_cls(
    klass: Type[ObjectT],
    *,
    delay: bool = False,
    strict: StrictModeT = STRICT_MODE,
    jsonschema: bool = True,
    serde: SerdeFlags = SerdeFlags(),
) -> Type[WrappedObjectT]:
    """Wrap a class to automatically enforce type-coercion on init.

    Notes
    -----
    While `Coercer.wrap` will work with classes alone, it changes the signature of the
    object to a function, there-by breaking inheritance. This follows a similar pattern to
    :func:`dataclasses.dataclasses`, which executes the function when wrapped, preserving
    the signature of the target class.

    Parameters
    ----------
    klass
        The class you wish to patch with coercion.
    delay
        Delay annotation resolution until first initialization.
    strict
        Turn on "validator mode": e.g. validate incoming data rather than coerce.
    jsonschema
        Generate a JSON Schema entry for this object.
    serde
        Optional settings for serialization/deserialization
    """

    def cls_wrapper(cls_: Type[ObjectT]) -> Type[WrappedObjectT]:
        setattr(cls_, "__delayed__", delay)
        if delay:
            _delay_resolve_class(
                cls_, strict=strict, jsonschema=jsonschema, serde=serde
            )
            return cast(Type[WrappedObjectT], cls_)

        return _resolve_class(cls_, strict=strict, jsonschema=jsonschema, serde=serde)

    wrapped: Type[WrappedObjectT] = cls_wrapper(klass)
    return wrapped


def _get_setter(cls: Type, bases: Tuple[Type, ...] = None):
    bases = bases or cls.__bases__
    setter = cls.__setattr__
    if setter.__name__ == "__setattr_coerced__":
        for base in bases:
            name = (
                ORIG_SETTER_NAME if hasattr(base, ORIG_SETTER_NAME) else "__setattr__"
            )
            setter = getattr(base, name, None)
            if setter.__name__ != "__setattr_coerced__":
                break
    return setter


def typed(_cls_or_callable=None, *, delay: bool = False, strict: bool = None):
    """A convenience function which automatically selects the correct wrapper.

    Parameters
    ----------
    delay
        Optionally delay annotation resolution until first call.
    strict
        Turn on "validator mode": e.g. validate incoming data rather than coerce.

    Returns
    -------
    The target object, appropriately wrapped.
    """
    strict = STRICT_MODE if strict is None else strict  # type: ignore

    def _typed(obj: Union[Callable, Type[ObjectT]]):
        if inspect.isclass(obj):
            return wrap_cls(obj, delay=delay, strict=strict)  # type: ignore
        elif isinstance(obj, Callable):  # type: ignore
            return wrap(obj, delay=delay, strict=strict)  # type: ignore
        else:
            raise TypeError(
                f"{__name__} requires a callable or class. Provided: {type(obj)}: {obj}"
            )

    return _typed(_cls_or_callable) if _cls_or_callable is not None else _typed


def resolve():
    """Resolve any delayed annotations.

    If this is not called, annotations will be resolved on first call
    of the wrapped class or callable.

    Examples
    --------
    >>> import typic
    >>>
    >>> @typic.klass(delay=True)
    ... class Duck:
    ...     color: str
    ...
    >>> typic.resolve()
    """
    while _TO_RESOLVE:
        obj = _TO_RESOLVE.pop()
        protocols(obj)
        if inspect.isclass(obj):
            obj.resolve()


_CONSTRAINT_TYPE_MAP = {
    x.type: x
    for x in c.ConstraintsT.__args__  # type: ignore
    if hasattr(x, "type") and x.type != object
}


def _get_constraint_cls(cls: Type) -> Optional[Type[c.ConstraintsT]]:
    if cls in _CONSTRAINT_TYPE_MAP:  # pragma: nocover
        return _CONSTRAINT_TYPE_MAP[cls]
    for typ, constr in _CONSTRAINT_TYPE_MAP.items():
        if issubclass(origin(cls), typ):
            _CONSTRAINT_TYPE_MAP[cls] = constr
            return constr

    return None


def _get_maybe_multi_constraints(
    v,
) -> Union[c.ConstraintsProtocolT, Tuple[c.ConstraintsProtocolT, ...]]:
    cons: Union[c.ConstraintsProtocolT, Tuple[c.ConstraintsProtocolT, ...]]
    if isinstance(v, Iterable):
        cons_gen = (c.get_constraints(x) for x in v)
        cons = tuple((x for x in cons_gen if x is not None))
    else:
        cons = c.get_constraints(v)
    return cons


def _handle_constraint_values(constraints, values, args):
    vcons = _get_maybe_multi_constraints(values)
    if vcons:
        constraints["values"] = vcons
        if not isinstance(values, tuple) or len(values) == 1:
            args = values if isinstance(values, tuple) else (values,)  # type: ignore
    return args


def _handle_constraint_keys(constraints, args):
    if "keys" in constraints:
        ks = constraints["keys"]
        kcons = _get_maybe_multi_constraints(ks)
        if kcons:
            constraints["keys"] = kcons
        if not isinstance(ks, tuple) or len(ks) == 1:
            args = (ks if isinstance(ks, tuple) else (ks,)) + args
            if len(args) == 1:
                args = args + (Any,)
    if len(args) == 1:
        args = (Any,) + args
    for k in ("items", "patterns", "key_dependencies"):
        if k in constraints:
            f = {}
            for n, c_ in constraints[k].items():
                c__cons = _get_maybe_multi_constraints(c_)
                if c__cons:
                    f[n] = c__cons

            constraints[k] = FrozenDict(f)
    return args


def constrained(
    _klass=None, *, values: Union[Type, Tuple[Type, ...]] = None, **constraints
):
    """A wrapper to indicate a 'constrained' type.

    Parameters
    ----------
    values
        For container-types, you can pass in other constraints for the values to be
        validated against. Can be a single constraint for all values or a tuple of
        constraints to choose from.

    **constraints
        The restrictions to apply to values being cast as the decorated type.

    Examples
    --------
    >>> import typic
    >>>
    >>> @typic.constrained(max_length=10)
    ... class ShortStr(str):
    ...     '''A short string.'''
    ...     ...
    ...
    >>> ShortStr('foo')
    'foo'
    >>> ShortStr('waytoomanycharacters')
    Traceback (most recent call last):
    ...
    typic.constraints.error.ConstraintValueError: Given value <'waytoomanycharacters'> fails constraints: (type=str, nullable=False, max_length=10)
    >>> @typic.constrained(values=ShortStr, max_items=2)
    ... class SmallMap(dict):
    ...     '''A small map that only allows short strings.'''
    ...
    >>> import json
    >>> print(json.dumps(typic.schema(SmallMap, primitive=True), indent=2))
    {
      "type": "object",
      "title": "SmallMap",
      "description": "A small map that only allows short strings.",
      "additionalProperties": {
        "type": "string",
        "maxLength": 10
      },
      "maxProperties": 2
    }


    See Also
    --------
    :py:mod:`typic.constraints.array`

    :py:mod:`typic.constraints.common`

    :py:mod:`typic.constraints.error`

    :py:mod:`typic.constraints.mapping`

    :py:mod:`typic.constraints.number`

    :py:mod:`typic.constraints.text`
    """

    def constr_wrapper(cls_: Type[ObjectT]) -> Type[ObjectT]:
        nonlocal constraints
        nonlocal values
        cdict = dict(cls_.__dict__)
        cdict.pop("__dict__", None)
        cdict.pop("__weakref__", None)
        constr_cls = _get_constraint_cls(cls_)
        if not constr_cls:
            raise TypeError(f"can't constrain type {cls_.__name__!r}")

        args: Tuple[Type, ...] = ()
        if values and constr_cls.type in {list, dict, set, tuple, frozenset}:  # type: ignore
            args = _handle_constraint_values(constraints, values, args)
        if constr_cls.type == dict:  # type: ignore
            args = _handle_constraint_keys(constraints, args)
        if args:
            cdict["__args__"] = args

        constraints_inst = constr_cls(**constraints)
        bases = inspect.getmro(cls_)

        def new(_new):
            @functools.wraps(_new)
            def __constrained_new(*args, **kwargs):
                result = _new(*args, **kwargs)
                return constraints_inst.validate(result)

            return __constrained_new

        def init(_init):
            @functools.wraps(_init)
            def __constrained_init(self, *args, **kwargs):
                _init(self, *args, **kwargs)
                constraints_inst.validate(self)

            return __constrained_init

        cdict.update(
            __constraints__=constraints_inst,
            __parent__=constraints_inst.type,
            **(
                {"__new__": new(cls_.__new__)}
                if constraints_inst.type in {str, bytes, int, float}
                else {"__init__": init(cls_.__init__)}
            ),
        )
        cls: Type[ObjectT] = cast(Type[ObjectT], type(cls_.__name__, bases, cdict))

        return cls

    return constr_wrapper(_klass) if _klass else constr_wrapper


environ = Environ(resolver)


def settings(
    _klass=None,
    *,
    prefix: str = "",
    case_sensitive: bool = False,
    frozen: bool = True,
    aliases: Mapping = None,
):
    """Create a typed class which fetches its defaults from env vars.

    The resolution order of values is `default(s) -> env value(s) -> passed value(s)`.

    Settings instances are indistinguishable from other `typical` dataclasses at
    run-time and are frozen by default. If you really want your settings to be mutable,
    you may pass in `frozen=False` manually.

    Parameters
    ----------
    prefix
        The prefix to strip from you env variables, i.e., `APP_`
    case_sensitive
        Whether your variables are case-sensitive. Defaults to `False`.
    frozen
        Whether to generate a frozen dataclass. Defaults to `True`
    aliases
        An optional mapping of potential aliases for your dataclass's fields.
        `{'other_foo': 'foo'}` will locate the env var `OTHER_FOO` and place it
        on the `Bar.foo` attribute.

    Examples
    --------
    >>> import os
    >>> import typic
    >>>
    >>> os.environ['FOO'] = "1"
    >>>
    >>> @typic.settings
    ... class Bar:
    ...     foo: int
    ...
    >>> Bar()
    Bar(foo=1)
    >>> Bar("3")
    Bar(foo=3)
    >>> bar = Bar()
    >>> bar.foo = 2
    Traceback (most recent call last):
    ...
    dataclasses.FrozenInstanceError: cannot assign to field 'foo'
    """
    aliases = aliases or {}

    def settings_wrapper(_cls):
        _resolve_from_env(_cls, prefix, case_sensitive, aliases)
        cls = wrap_cls(dataclasses.dataclass(_cls, frozen=frozen), jsonschema=False)
        return cls

    return settings_wrapper(_klass) if _klass is not None else settings_wrapper


def _resolve_from_env(
    cls: Type[ObjectT],
    prefix: str,
    case_sensitive: bool,
    aliases: Mapping[str, str],
) -> Type[ObjectT]:
    fields = cached_type_hints(cls)
    vars = {
        (f"{prefix}{x}".lower() if not case_sensitive else f"{prefix}{x}"): (x, y)
        for x, y in fields.items()
    }
    attr_to_aliases = collections.defaultdict(set)
    for alias, attr in aliases.items():
        attr_to_aliases[attr].add(alias)

    sentinel = object()
    for name in vars:
        attr, typ = vars[name]
        names = attr_to_aliases[name]
        field = getattr(cls, attr, sentinel)
        if field is sentinel:
            field = dataclasses.field()
        elif not isinstance(field, dataclasses.Field):
            field = dataclasses.field(default=field)
        if field.default_factory != dataclasses.MISSING:
            continue

        kwargs = dict(var=name, ci=not case_sensitive)
        if field.default != dataclasses.MISSING:
            kwargs["default"] = field.default
            field.default = dataclasses.MISSING

        factory = environ.register(typ, *names, name=name)
        field.default_factory = functools.partial(factory, **kwargs)
        setattr(cls, attr, field)

    return cls


PrimitiveT = Union[Dict, List, str, int, bool]
SchemaPrimitiveT = Dict[str, PrimitiveT]
SchemaReturnT = Union[SchemaPrimitiveT, ObjectSchemaField]


@lru_cache(maxsize=None)
def schema(obj: Type[ObjectT], *, primitive: bool = False) -> SchemaReturnT:
    """Get a JSON schema for object for the given object.

    Parameters
    ----------
    obj
        The class for which you wish to generate a JSON schema
    primitive
        Whether to return an instance of :py:class:`typic.schema.ObjectSchemaField` or
        a "primitive" (dict object).

    Examples
    --------
    >>> import typic
    >>>
    >>> @typic.klass
    ... class Foo:
    ...     bar: str
    ...
    >>> typic.schema(Foo)
    ObjectSchemaField(title='Foo', description='Foo(bar: str)', properties={'bar': StrSchemaField()}, additionalProperties=False, required=('bar',))
    >>> typic.schema(Foo, primitive=True)
    {'type': 'object', 'title': 'Foo', 'description': 'Foo(bar: str)', 'properties': {'bar': {'type': 'string'}}, 'additionalProperties': False, 'required': ['bar'], 'definitions': {}}

    """
    if obj in {FunctionType, MethodType}:
        raise ValueError("Cannot build schema for function or method.")

    annotation = resolver.resolve(obj)
    schm = schema_builder.get_field(annotation)
    try:
        setattr(obj, SCHEMA_NAME, schm)
    except (AttributeError, TypeError):
        pass
    return cast(SchemaReturnT, schm.primitive() if primitive else schm)



from __future__ import annotations

import dataclasses
import functools
import inspect
import warnings
from enum import Enum
from operator import attrgetter, methodcaller
from typing import (
    Mapping,
    Any,
    Type,
    Optional,
    ClassVar,
    Tuple,
    cast,
    Union,
    TypeVar,
    Iterator,
    Dict,
    Iterable,
)

from typic import checks, constraints as constr, util, strict as st
from typic.common import (
    EMPTY,
    ORIG_SETTER_NAME,
    SERDE_FLAGS_ATTR,
    TYPIC_ANNOS_NAME,
    ObjectT,
    Case,
    ReadOnly,
)
from typic.compat import lru_cache
from typic.ext import json
from typic.strict import StrictModeT
from .binder import Binder
from .common import (
    SerializerT,
    SerdeFlags,
    SerdeConfig,
    Annotation,
    SerdeProtocol,
    SerdeProtocolsT,
    DelayedSerdeProtocol,
    ForwardDelayedAnnotation,
    DelayedAnnotation,
    DeserializerT,
    EncoderT,
    DecoderT,
    TranslatorT,
    PrimitiveT,
    FieldIteratorT,
)
from .des import DesFactory
from .ser import SerFactory
from .translator import TranslatorFactory

_T = TypeVar("_T")


class Resolver:
    """A type serializer/deserializer resolver."""

    STRICT = st.STRICT_MODE
    _DICT_FACTORY_METHODS = frozenset(
        {("asdict", methodcaller("asdict")), ("to_dict", methodcaller("to_dict"))}
    )
    _DYNAMIC = SerFactory._DYNAMIC
    OPTIONALS = (None, ...)
    LITERALS = (int, bytes, str, bool, Enum, type(None))

    def __init__(self):
        self.des = DesFactory(self)
        self.ser = SerFactory(self)
        self.binder = Binder(self)
        self.translator = TranslatorFactory(self)
        self.bind = self.binder.bind
        self.__cache = {}
        self.__stack = set()
        for typ in checks.STDLIB_TYPES:
            self.resolve(typ)
            self.resolve(Optional[typ])
            self.resolve(typ, is_optional=True)
            try:
                self.translator.iterator(typ)
                self.translator.iterator(typ, values=True)
            except TypeError:
                pass

    def transmute(self, annotation: Type[ObjectT], value: Any) -> ObjectT:
        """Convert a given value `into` the target annotation.

        Checks for:
            - :class:`datetime.date`
            - :class:`datetime.datetime`
            - builtin types
            - extended type annotations as described in the ``typing`` module.
            - User-defined classes (limited)

        Parameters
        ----------
        annotation :
            The provided annotation for determining the coercion
        value :
            The value to be transmuted
        """
        resolved: SerdeProtocol = self.resolve(annotation)
        transmuted: ObjectT = resolved.transmute(value)

        return transmuted

    def translate(self, value: ObjectT, target: Type[_T]) -> _T:
        """Translate an instance `from` its type `to` a target type.

        Notes
        -----
        This provides a functional interface for translating one custom class
        instance to another custom class. This should not be confused with
        :py:func:`typic.transmute`, which is generally a more powerful functional
        interface for conversion between types, but this is provided as for
        api-completeness with the object-api.

        Parameters
        ----------
        value
            The higher-order class instance to translate.
        target
            The higher-order class to translate into.
        """
        t = value.__class__
        resolved: SerdeProtocol = self.resolve(t)
        return resolved.translate(value, target)

    def validate(
        self, annotation: Type[ObjectT], value: Any, *, transmute: bool = False
    ) -> Union[ObjectT, Any]:
        """Validate an input against the type-constraints for the given annotation.

        Parameters
        ----------
        annotation
            The type or annotation to validate against
        value
            The value to check
        transmute: (kw-only)
            Whether to transmute the value to the annotation after validation
        """
        resolved: SerdeProtocol = self.resolve(annotation)
        value = resolved.validate(value)
        if transmute:
            return resolved.transmute(value)
        return value

    def iterate(
        self, obj, *, values: bool = False, exclude: Iterable[str] = ()
    ) -> Iterator[Union[Tuple[str, Any], Any]]:
        """Iterate over the fields of an object.

        Parameters
        ----------
        obj
            The object to iterate over
        values
            Whether to only yield values of an object's fields. (defaults False)
        exclude
            Proactively ignore any fields on the object
        """
        t = obj.__class__
        # Extract the type of the enum value if this is an Enum.
        # Enums classes are iterable and will generate the wrong kind of iterator.
        if checks.isenumtype(t):
            obj = obj.value
            t = obj.__class__
        iterator = self.translator.iterator(t, values=values, exclude=(*exclude,))
        return iterator(obj)

    def coerce_value(
        self, value: Any, annotation: Type[ObjectT]
    ) -> ObjectT:  # pragma: nocover
        warnings.warn(
            "'typic.coerce' has been deprecated and will be removed in a future "
            "version. Use 'typic.transmute' instead.",
            DeprecationWarning,
            stacklevel=3,
        )
        return self.transmute(annotation, value)

    def known(self, t: Type) -> bool:
        return hasattr(t, ORIG_SETTER_NAME) or hasattr(t, "__delayed__")

    def delayed(self, t: Type) -> bool:
        return getattr(t, "__delayed__", False)

    def primitive(
        self,
        obj: ObjectT,
        *,
        lazy: bool = False,
        name: util.ReprT = None,
        flags: SerdeFlags = None,
    ) -> PrimitiveT:
        """A method for converting an object to its primitive equivalent.

        Useful for encoding data to JSON.

        Examples
        --------
        >>> import typic
        >>> import datetime
        >>> import uuid
        >>> import ipaddress
        >>> import re
        >>> import dataclasses
        >>> typic.primitive("foo")
        'foo'
        >>> typic.primitive(("foo",))  # containers are converted to lists/dicts
        ['foo']
        >>> typic.primitive(datetime.datetime(1970, 1, 1))
        '1970-01-01T00:00:00'
        >>> typic.primitive(b"foo")
        'foo'
        >>> typic.primitive(ipaddress.IPv4Address("0.0.0.0"))
        '0.0.0.0'
        >>> typic.primitive(re.compile("[0-9]"))
        '[0-9]'
        >>> typic.primitive(uuid.UUID(int=0))
        '00000000-0000-0000-0000-000000000000'
        >>> @dataclasses.dataclass
        ... class Foo:
        ...     bar: str = 'bar'
        ...
        >>> typic.primitive(Foo())
        {'bar': 'bar'}
        """
        t = obj.__class__
        if checks.isenumtype(t):
            obj = obj.value  # type: ignore
            t = obj.__class__
        proto: SerdeProtocol = self.resolve(t, flags=flags)
        return proto.primitive(obj, lazy=lazy, name=name)  # type: ignore

    def tojson(
        self, obj: ObjectT, *, indent: int = 0, ensure_ascii: bool = False, **kwargs
    ) -> str:
        """A method for dumping any object to a valid JSON string.

        Notes
        -----
        If `ujson` is installed, we will default to that library for the final
        encoding, which can result in massive performance gains over the standard `json`
        library.

        Examples
        --------
        >>> import typic
        >>> import datetime
        >>> import uuid
        >>> import ipaddress
        >>> import re
        >>> import dataclasses
        >>> import enum
        >>> typic.tojson("foo")
        '"foo"'
        >>> typic.tojson(("foo",))
        '["foo"]'
        >>> typic.tojson(datetime.datetime(1970, 1, 1))
        '"1970-01-01T00:00:00"'
        >>> typic.tojson(b"foo")
        '"foo"'
        >>> typic.tojson(ipaddress.IPv4Address("0.0.0.0"))
        '"0.0.0.0"'
        >>> typic.tojson(re.compile("[0-9]"))
        '"[0-9]"'
        >>> typic.tojson(uuid.UUID(int=0))
        '"00000000-0000-0000-0000-000000000000"'
        >>> @dataclasses.dataclass
        ... class Foo:
        ...     bar: str = 'bar'
        ...
        >>> typic.tojson(Foo())
        '{"bar":"bar"}'
        >>> class Enum(enum.Enum):
        ...     FOO = "foo"
        ...
        >>> typic.tojson(Enum.FOO)
        '"foo"'
        """
        t = obj.__class__
        if checks.isenumtype(t):
            obj = obj.value  # type: ignore
            t = obj.__class__
        proto: SerdeProtocol = self.resolve(t)
        return proto.tojson(obj, indent=indent, ensure_ascii=ensure_ascii, **kwargs)

    def decode(
        self, annotation: Type[ObjectT], value: Any, decoder: DecoderT[bytes], **kwargs
    ) -> ObjectT:
        proto: SerdeProtocol = self.resolve(annotation)
        return proto.transmute(decoder(value, **kwargs))  # type: ignore

    def encode(self, obj: Any, encoder: EncoderT[PrimitiveT], **kwargs) -> bytes:
        t = obj.__class__
        if checks.isenumtype(t):
            obj = obj.value
            t = obj.__class__
        proto: SerdeProtocol = self.resolve(t)
        return encoder(proto.primitive(obj), **kwargs)  # type: ignore

    @lru_cache(maxsize=None)
    def _get_configuration(self, origin: Type, flags: SerdeFlags) -> SerdeConfig:
        if hasattr(origin, SERDE_FLAGS_ATTR):
            flags = getattr(origin, SERDE_FLAGS_ATTR)
        # Get all the annotated fields
        params = util.safe_get_params(origin)
        # This is probably a builtin and has no signature
        fields: Dict[str, Annotation] = {}
        hints = util.cached_type_hints(origin)
        for name, t in hints.items():
            fields[name] = self.annotation(
                t,
                flags=dataclasses.replace(flags, fields={}),
                default=getattr(origin, name, EMPTY),
                namespace=origin,
            )

        # Filter out any annotations which aren't part of the object's signature.
        if flags.signature_only:
            fields = {x: fields[x] for x in fields.keys() & params.keys()}
        # Create a field-to-field mapping
        fields_out = {x: x for x in fields}
        # Make sure to include any fields explicitly listed
        include = flags.fields
        if include:
            if isinstance(include, Mapping):
                fields_out.update(include)
            else:
                fields_out.update({x: x for x in include})
        # Transform the output fields to the correct case.
        if flags.case:
            case = Case(flags.case)
            fields_out = {x: case.transformer(y) for x, y in fields_out.items()}
        omit = flags.omit
        # Omit fields with explicitly omitted types & flag values to omit at dump
        value_omissions: Tuple[Any, ...] = ()
        if omit:
            type_omissions = {
                o for o in omit if checks._type_check(o) or o is NotImplemented
            }
            type_name_omissions = {util.get_name(o) for o in type_omissions}
            value_omissions = (*(o for o in omit if o not in type_omissions),)
            fields_out_final = {}
            anno: Union[Annotation, DelayedAnnotation, ForwardDelayedAnnotation]
            for name, out in fields_out.items():
                anno = fields[name]
                default = anno.parameter.default if anno.parameter else EMPTY
                if isinstance(anno, ForwardDelayedAnnotation):
                    if (
                        not {util.get_name(anno.ref), util.get_name(default)}
                        & type_name_omissions
                    ):
                        fields_out_final[name] = out
                elif not {anno.origin, default} & type_omissions:
                    fields_out_final[name] = out
            fields_out = fields_out_final

        fields_in = {y: x for x, y in fields_out.items()}
        if params:
            fields_in = {x: y for x, y in fields_in.items() if y in params}
        exclude = flags.exclude
        if exclude:
            fields_out = {x: y for x, y in fields_out.items() if x not in exclude}
        fields_getters = {x: attrgetter(x) for x in fields}
        return SerdeConfig(
            flags=flags,
            fields=fields,
            fields_out=fields_out,
            fields_in=fields_in,
            fields_getters=fields_getters,
            omit_values=value_omissions,
            encoder=flags.encoder,
            decoder=flags.decoder,
        )

    def annotation(
        self,
        annotation: Type[ObjectT],
        name: str = None,
        parameter: Optional[inspect.Parameter] = None,
        is_optional: bool = None,
        is_strict: StrictModeT = None,
        flags: "SerdeFlags" = None,
        default: Any = EMPTY,
        namespace: Type = None,
    ) -> Annotation[Type[ObjectT]]:
        """Get a :py:class:`Annotation` for this type.

        Unlike a :py:class:`ResolvedAnnotation`, this does not provide access to a
        serializer/deserializer/validator protocol.
        """
        flags = cast(
            "SerdeFlags", getattr(annotation, SERDE_FLAGS_ATTR, flags or SerdeFlags())
        )
        if parameter is None:
            parameter = inspect.Parameter(
                name or "_",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=annotation,
                default=default if checks.ishashable(default) else ...,
            )
        # Check for the super-type
        non_super = util.resolve_supertype(annotation)
        # Note, this may be a generic, like Union.
        orig = util.origin(annotation)
        use = non_super
        # Get the unfiltered args
        args = getattr(non_super, "__args__", None)
        # Set whether this is optional/strict
        is_optional = (
            is_optional
            or checks.isoptionaltype(non_super)
            or parameter.default in self.OPTIONALS
        )
        is_strict = is_strict or checks.isstrict(non_super) or self.STRICT
        is_static = util.origin(use) not in self._DYNAMIC
        is_literal = checks.isliteral(use)
        # Determine whether we should use the first arg of the annotation
        while checks.should_unwrap(use) and args:
            is_optional = is_optional or checks.isoptionaltype(use)
            is_strict = is_strict or checks.isstrict(use)
            if is_optional and len(args) > 2:
                # We can't resolve this annotation.
                is_static = False
                use = Union[args[:-1]]
                break
            # Note that we don't re-assign `orig`.
            # This is intentional.
            # Special forms are needed for building the downstream validator.
            # Callers should be aware of this and perhaps use `util.origin` elsewhere.
            non_super = util.resolve_supertype(args[0])
            use = non_super
            args = util.get_args(use)
            is_static = util.origin(use) not in self._DYNAMIC
            is_literal = is_literal or checks.isliteral(use)

        # Only allow legal parameters at runtime, this has implementation implications.
        if is_literal:
            args = util.get_args(use)
            if any(not isinstance(a, self.LITERALS) for a in args):
                raise TypeError(
                    f"PEP 586: Unsupported parameters for 'Literal' type: {args}. "
                    "See https://www.python.org/dev/peps/pep-0586/"
                    "#legal-parameters-for-literal-at-type-check-time "
                    "for more information."
                )
        # The type definition doesn't exist yet.
        if checks.isforwardref(use):
            module = self.__module__
            # Ideally we have a namespace from a parent class/function to the field
            if namespace:
                module = namespace.__module__

            fda = ForwardDelayedAnnotation(
                ref=use,
                resolver=self,
                _name=name,
                parameter=parameter,
                is_optional=is_optional,
                is_strict=is_strict,
                flags=flags,
                default=default,
                module=module,
                frame=inspect.currentframe(),
            )
            return cast(Annotation, fda)
        # The type definition is recursive or within a recursive loop.
        elif use is namespace or use in self.__stack:
            # If detected via stack, we can remove it now.
            # Otherwise we'll cause another recursive loop.
            if use in self.__stack:
                self.__stack.remove(use)
            da = DelayedAnnotation(
                type=use,
                resolver=self,
                _name=name,
                parameter=parameter,
                is_optional=is_optional,
                is_strict=is_strict,
                flags=flags,
                default=default,
            )
            return cast(Annotation, da)
        # Otherwise, add this type to the stack to prevent a recursive loop from elsewhere.
        if not checks.isstdlibtype(use):
            self.__stack.add(use)
        serde = (
            self._get_configuration(util.origin(use), flags)
            if is_static and not is_literal
            else SerdeConfig(flags)
        )

        anno: Annotation = Annotation(
            resolved=use,
            origin=orig,
            un_resolved=annotation,
            parameter=parameter,
            optional=is_optional,
            strict=is_strict,
            static=is_static,
            serde=serde,
        )
        anno.translator = cast(
            TranslatorT, functools.partial(self.translator.factory, anno)
        )
        return anno

    def _resolve_from_annotation(
        self,
        anno: Annotation[Type[ObjectT]],
        *,
        namespace: Type = None,
    ) -> SerdeProtocol[ObjectT]:
        if anno in self.__cache:
            return self.__cache[anno]
        if isinstance(anno, (DelayedAnnotation, ForwardDelayedAnnotation)):
            return DelayedSerdeProtocol(anno)

        # Build the deserializer
        constraints = constr.get_constraints(
            anno.resolved, nullable=anno.optional, cls=namespace
        )
        deserializer, validator = self.des.factory(
            anno, constraints, namespace=namespace
        )
        # Build the serializer
        serializer = self.ser.factory(anno)
        # Put it all together
        proto = self._build_protocol(
            annotation=anno,
            constraints=constraints,
            deserializer=deserializer,
            validator=validator,
            serializer=serializer,
        )
        self.__cache[anno] = proto
        return proto

    def _build_protocol(
        self,
        *,
        annotation: Annotation[Type[ObjectT]],
        constraints: constr.ConstraintsProtocolT[ObjectT],
        deserializer: DeserializerT[ObjectT],
        validator: constr.ValidateT[ObjectT],
        serializer: SerializerT[ObjectT],
    ) -> SerdeProtocol[ObjectT]:
        def tojson(
            val: ObjectT,
            *,
            indent: int = 0,
            ensure_ascii: bool = False,
            __prim=serializer,
            __dumps=json.dumps,
            **kwargs,
        ) -> str:
            return __dumps(
                __prim(val),
                indent=indent,
                ensure_ascii=ensure_ascii,
                **kwargs,
            )

        tojson.__qualname__ = f"{SerdeProtocol.__name__}.{tojson.__name__}"
        tojson.__module__ = SerdeProtocol.__module__

        # Set the encoder/decoder protocols.
        # Default to JSON for wire-format
        encode: EncoderT = cast(EncoderT, tojson)
        if annotation.serde.encoder:

            def encode(  # type: ignore
                val: ObjectT,
                *,
                __prim=serializer,
                __encode=annotation.serde.encoder,
                **kwargs,
            ) -> bytes:
                return __encode(__prim(val), **kwargs)

            encode.__qualname__ = f"{SerdeProtocol.__name__}.{encode.__name__}"
            encode.__module__ = self.__class__.__module__

        # Default to JSON for wire-format
        decode: DecoderT = cast(DecoderT, deserializer)
        if annotation.serde.decoder:

            def decode(  # type: ignore
                val: bytes,
                *,
                __trans=deserializer,
                __decode=annotation.serde.decoder,
                **kwargs,
            ) -> ObjectT:
                return __trans(__decode(val, **kwargs))

            decode.__qualname__ = f"{SerdeProtocol.__name__}.{decode.__name__}"
            decode.__module__ = SerdeProtocol.__module__

        # Create the translator
        def translate(
            value: ObjectT, target: Type[_T], *, __factory=annotation.translator
        ) -> _T:
            trans = __factory(target)
            return trans(value)

        translate.__qualname__ = f"{SerdeProtocol.__name__}.{translate.__name__}"
        translate.__module__ = SerdeProtocol.__module__

        # Create the iterator, if possible.
        try:
            iterator = self._iterator_from_annotation(annotation)
            iterator.__qualname__ = f"{SerdeProtocol.__name__}.{iterator.__name__}"
            iterator.__module__ = SerdeProtocol.__module__
        # Default to lazy iteration, if not.
        except TypeError:
            iterator = cast(FieldIteratorT, self.iterate)

        return SerdeProtocol(
            annotation=annotation,
            constraints=constraints,
            deserialize=deserializer,
            decode=decode,
            serialize=serializer,
            encode=encode,
            validate=validator,
            translate=cast(TranslatorT, translate),
            tojson=cast(EncoderT, tojson),
            iterate=iterator,
        )

    def _iterator_from_annotation(
        self, annotation: Annotation[Type[ObjectT]]
    ) -> FieldIteratorT[ObjectT]:
        exclude = (*(annotation.serde.flags.exclude or ()),)
        fiterator = self.translator.iterator(
            annotation.resolved_origin, relaxed=True, exclude=exclude
        )
        viterator = self.translator.iterator(
            annotation.resolved_origin,
            values=True,
            relaxed=True,
            exclude=exclude,
        )

        def iterator(
            o: ObjectT, *, values: bool = False, __fields=fiterator, __values=viterator
        ) -> Iterator[Union[Tuple[str, Any], Any]]:
            if values:
                return __values(o)
            return __fields(o)

        return cast(FieldIteratorT, iterator)

    @lru_cache(maxsize=None)
    def resolve(
        self,
        annotation: Type[ObjectT],
        *,
        flags: SerdeFlags = None,
        name: str = None,
        parameter: Optional[inspect.Parameter] = None,
        is_optional: bool = None,
        is_strict: bool = None,
        namespace: Type = None,
    ) -> SerdeProtocol[ObjectT]:
        """Get a :py:class:`SerdeProtocol` from a given annotation or type.

        Parameters
        ----------
        annotation
            The class or callable object you wish to extract resolved annotations from.

        Other Parameters
        ----------------
        flags : (optional)
            An instance of :py:class:`SerdeFlags`
        name : (optional)
            An name, such as an attribute or parameter name.
        parameter: (optional)
            The parameter associated to this annotation, if any.
        is_optional: (optional)
            Whether to allow null values.
        is_strict: (optional)
            Whether to apply strict validation to any input for this annotation.

        Examples
        --------
        >>> import typic
        >>>
        >>> @typic.klass
        ... class Foo:
        ...     bar: str
        ...
        >>> protocol = typic.protocol(Foo)

        See Also
        --------
        :py:class:`SerdeProtocol`
        """
        # Extract the meta-data.
        anno = self.annotation(
            annotation=annotation,
            name=name,
            parameter=parameter,
            is_optional=is_optional,
            is_strict=is_strict,
            flags=flags,
            namespace=namespace,
        )
        resolved = self._resolve_from_annotation(anno, namespace=namespace)
        self.__stack.clear()
        return resolved

    @lru_cache(maxsize=None)
    def protocols(self, obj, *, strict: bool = False) -> SerdeProtocolsT:
        """Get a mapping of param/attr name -> :py:class:`SerdeProtocol`

        Parameters
        ----------
        obj
            The class or callable object you wish to extract resolved annotations from.
        strict
            Whether to validate instead of coerce.

        Examples
        --------
        >>> import typic
        >>>
        >>> @typic.klass
        ... class Foo:
        ...     bar: str
        ...
        >>> protocols = typic.protocols(Foo)

        See Also
        --------
        :py:class:`SerdeProtocol`
        """

        if not any(
            (inspect.ismethod(obj), inspect.isfunction(obj), inspect.isclass(obj))
        ):
            obj = obj.__class__

        hints = util.cached_type_hints(obj)
        params = util.safe_get_params(obj)
        fields: Mapping[str, dataclasses.Field] = {}
        if dataclasses.is_dataclass(obj):
            fields = {f.name: f for f in dataclasses.fields(obj)}
        ann = {}
        for name in params.keys() | hints.keys():
            param = params.get(name)
            hint = hints.get(name)
            field = fields.get(name)
            annotation = hint or param.annotation  # type: ignore
            annotation = util.resolve_supertype(annotation)
            param = param or inspect.Parameter(
                name,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=EMPTY,
                annotation=hint or annotation,
            )
            if repr(param.default) == "<factory>":
                param = param.replace(default=EMPTY)
            if checks.isclassvartype(annotation):
                val = getattr(obj, name)
                if annotation is ClassVar:
                    annotation = annotation[type(val)]
                default = val
                param = param.replace(default=default)
            if (
                field
                and field.default is not dataclasses.MISSING
                and param.default is EMPTY
            ):
                if field.init is False and util.origin(annotation) is not ReadOnly:
                    annotation = ReadOnly[annotation]  # type: ignore
                param = param.replace(default=field.default)

            if not checks.ishashable(param.default):
                param = param.replace(default=...)

            resolved = self.resolve(
                annotation, name=name, parameter=param, is_strict=strict, namespace=obj
            )
            ann[name] = resolved
        try:
            setattr(obj, TYPIC_ANNOS_NAME, ann)
        # We wrapped a bound method, or
        # are wrapping a static-/classmethod
        # after they were wrapped with @static/class
        except (AttributeError, TypeError):
            pass

        return ann


resolver = Resolver()



from __future__ import annotations

import enum
import inspect
from functools import partial
from typing import Union, Type, Any, TypeVar, Callable, Generic, Mapping

import inflection

DEFAULT_ENCODING = "utf-8"
EMPTY = inspect.Signature.empty
ORIG_SETTER_NAME = "__setattr_original__"
POSITIONAL_ONLY = inspect.Parameter.POSITIONAL_ONLY
POSITIONAL_OR_KEYWORD = inspect.Parameter.POSITIONAL_OR_KEYWORD
KEYWORD_ONLY = inspect.Parameter.KEYWORD_ONLY
RETURN_KEY = "return"
SCHEMA_NAME = "__json_schema__"
SELF_NAME = "self"
SERDE_ATTR = "__serde__"
SERDE_FLAGS_ATTR = "__serde_flags__"
TOO_MANY_POS = "too many positional arguments"
TYPIC_ANNOS_NAME = "__typic_annotations__"
VAR_POSITIONAL = inspect.Parameter.VAR_POSITIONAL
VAR_KEYWORD = inspect.Parameter.VAR_KEYWORD
KWD_KINDS = {VAR_KEYWORD, KEYWORD_ONLY}
POS_KINDS = {VAR_POSITIONAL, POSITIONAL_ONLY}
AnyOrTypeT = Union[Type, Any]
ObjectT = TypeVar("ObjectT")
"""A generic alias for an object."""
OriginT = TypeVar("OriginT")
"""A type alias for an instance of the type associated to a Coercer."""
CaseTransformerT = Callable[[str], str]
"""A callable which transforms a string from one case-style to another."""


class Case(str, enum.Enum):
    """An enumeration of the supported case-styles for field names."""

    CAMEL = "camelCase"
    SNAKE = "snake_case"
    PASCAL = "PascalCase"
    KEBAB = "kebab-case"
    DOT = "dot.case"
    UPPER_KEBAB = "UPPER-KEBAB-CASE"
    UPPER_DOT = "UPPER.DOTA.CASE"

    @property
    def transformer(self) -> CaseTransformerT:
        return _TRANSFORMERS[self]


def upper_kebab_case(s: str) -> str:
    return inflection.dasherize(s).upper()


def upper_dot_case(s: str) -> str:
    return inflection.parameterize(s, separator=".").upper()


_TRANSFORMERS: Mapping[Case, CaseTransformerT] = {
    Case.CAMEL: partial(inflection.camelize, uppercase_first_letter=False),
    Case.SNAKE: inflection.underscore,
    Case.PASCAL: inflection.camelize,
    Case.KEBAB: inflection.dasherize,
    Case.DOT: partial(inflection.parameterize, separator="."),
    Case.UPPER_KEBAB: upper_kebab_case,
    Case.UPPER_DOT: upper_dot_case,
}
T = TypeVar("T")


class ReadOnly(Generic[T]):
    """A type annotation to indicate a field is meant to be read-only."""

    pass


class WriteOnly(Generic[T]):
    """A type annotation to indicate a field is meant to be write-only."""

    pass



from __future__ import annotations
import dataclasses
import enum
import inspect
import linecache
import pathlib
import uuid
from typing import List, Union, Type, Tuple, Optional, TypeVar

import typic
from .util import slotted

_empty = inspect.Parameter.empty
ParameterKind = inspect._ParameterKind


class rawstr(str):
    def __repr__(self):
        return super().__repr__().strip("'\"")


class Keyword(str, enum.Enum):
    RET = "return"
    YLD = "yield"
    ASN = "async"
    AWT = "await"
    DEF = "def"
    CLS = "class"
    AST = "assert"
    IMP = "import"
    FRM = "from"
    DEC = "@"


@slotted(dict=False)
@dataclasses.dataclass(frozen=True)
class Line:
    INDENT = "    "
    code: str
    level: int

    def render(self) -> str:
        if self.code.strip():
            return f"{self.INDENT * self.level}{self.code}\n"
        return self.code


_BT = TypeVar("_BT")


@slotted(dict=False)
@dataclasses.dataclass
class Block:
    namespace: dict = dataclasses.field(default_factory=dict)
    body: List[Union[Line, Block]] = dataclasses.field(default_factory=list)
    level: int = 0
    name: str = ""

    def line(self, line: str, *, level: int = None, **context):
        if level is None:
            level = self.level
        self.namespace.update(context)
        self.body.append(Line(line, level))

    l = line  # noqa: E741

    def block(self, *lines: str, level: int = None, name: str = "", **context) -> Block:
        level = (self.level + 1) if level is None else level
        self.namespace.update(context)
        block = Block(self.namespace, level=level, name=name)
        for line in lines:
            self.l(line)
        self.body.append(block)
        return block

    b = block

    @staticmethod
    def param(  # type: ignore
        name: str,
        kind: ParameterKind = ParameterKind.POSITIONAL_OR_KEYWORD,  # type: ignore
        *,
        default=_empty,
        annotation=_empty,
    ) -> inspect.Parameter:
        return inspect.Parameter(
            name=name, kind=kind, default=default, annotation=annotation
        )

    p = param

    def func(
        self,
        name: str,
        *params: inspect.Parameter,
        decorator: str = None,
        returns: Type = _empty,
        coro: bool = False,
        **context,
    ) -> Function:
        self.namespace.update(context)
        func = Function(
            namespace=self.namespace,
            level=self.level + 1,
            name=name,
            parameters=[*params],
            returns=returns,
            coro=coro,
            decorator=decorator,
        )
        self.body.append(func)
        return func

    f = func

    def cls(
        self, name: str, *params, base: str = None, decorator: str = None, **context
    ) -> Class:  # pragma: nocover
        self.namespace.update(context)
        cls = Class(
            namespace=self.namespace,
            level=self.level + 1,
            name=name,
            parameters=[*params],
            base=base,
            decorator=decorator,
        )
        self.body.append(cls)
        return cls

    c = cls

    def __enter__(self: _BT) -> _BT:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @staticmethod
    def _generate_unique_filename(func_name):
        """Create a "filename" suitable for a function being generated.

        Notes
        -----
        Taken approximately from `attrs`.

        See Also
        --------
        https://github.com/python-attrs/attrs/blob/8c00f755f9d91c06fbdd9a20e24d2c4663e6339d/src/attr/_make.py#L1066
        """
        unique_id = uuid.uuid4()
        extra = ""
        count = 1

        while True:
            unique_filename = f"<typical generated {func_name}{extra}>"
            # To handle concurrency we essentially "reserve" our spot in
            # the linecache with a dummy line.  The caller can then
            # set this value correctly.
            cache_line = (1, None, (str(unique_id),), unique_filename)
            if linecache.cache.setdefault(unique_filename, cache_line) == cache_line:
                return unique_filename

            # Looks like this spot is taken. Try again.
            count += 1
            extra = f"-{count}"

    def _add_to_linecache(self, fname, code):
        linecache.cache[fname] = (
            len(self.body),
            None,
            code.splitlines(True),
            fname,
        )

    def _render_head(self) -> str:
        return ""

    def _render_body(self) -> str:
        block = "".join(str(x.render()) for x in self.body)
        return block or "..."

    def render(self) -> str:
        return self._render_head() + self._render_body()

    def compile(self, *, name: str, ns: dict = None):
        ns = {} if ns is None else ns
        fname = self._generate_unique_filename(func_name=name)
        self.namespace.update(ns)
        code = self.render()
        bytecode = compile(code, fname, "exec")
        eval(bytecode, self.namespace, self.namespace)
        target = self.namespace[name]
        target.__raw__ = code
        self._add_to_linecache(fname, code)
        return target


@slotted(dict=False)
@dataclasses.dataclass
class Function(Block):
    parameters: List[inspect.Parameter] = dataclasses.field(default_factory=list)
    returns: Optional[Type] = None
    coro: bool = False
    decorator: Optional[str] = None

    def _render_head(self) -> str:
        defn = f"{Keyword.DEF} {self.name}"
        if self.coro:
            defn = f"{Keyword.ASN} {defn}"
        funcsig = f"{defn}{str(inspect.Signature(self.parameters))}:"
        lines = ["\n", funcsig, "\n"]
        if self.decorator:
            lines = ["\n", f"{Keyword.DEC}{self.decorator}", funcsig, "\n"]

        return "".join(lines)

    def localize_context(self, *context: str):
        """A byte-code hack to inject variables into the local namespace."""
        self.parameters.extend(
            (
                self.param(name=n, kind=ParameterKind.KEYWORD_ONLY, default=rawstr(n))
                for n in context
            )
        )

    def add_param(
        self,
        name: str,
        kind: ParameterKind = ParameterKind.POSITIONAL_OR_KEYWORD,  # type: ignore
        *,
        default=_empty,
        annotation=_empty,
    ):
        self.parameters.append(
            self.param(name=name, kind=kind, default=default, annotation=annotation)
        )


@slotted(dict=False)
@dataclasses.dataclass
class Class(Block):
    parameters: List[inspect.Parameter] = dataclasses.field(default_factory=list)
    decorator: Optional[str] = None
    base: Optional[str] = None

    def _render_head(self) -> str:
        decl = f"{Keyword.CLS} {self.name}"
        if self.base:
            decl = f"{decl}({self.base})"
        initfn = ""
        if self.parameters:
            sig = inspect.Signature(parameters=self.parameters)
            initfn = f"\n{Keyword.DEF} __init__{str(sig)}:"
        lines = ["\n", decl, "\n", initfn, "\n"]
        if self.decorator:
            lines = ["\n", f"{Keyword.DEC}{self.decorator}", *lines]
        return "".join(lines)


# This isn't used or tested. It's just here for API completion.
# Perhaps an optimization could be to allow users to pre-compile their protocols.
@dataclasses.dataclass
class Module:  # pragma: nocover
    namespace: dict = dataclasses.field(init=False, default_factory=dict)
    body: Block = dataclasses.field(init=False)

    def __post_init__(self):
        self.body = Block(self.namespace)

    @staticmethod
    def imports(namespace: dict) -> Tuple[Block, dict]:
        with Block() as imports:
            remains = {}
            for key, value in namespace.items():
                if inspect.ismodule(value):
                    imports.l(f"{Keyword.IMP} {value.__name__} as {key}", level=0)
                elif (
                    (inspect.isclass(value) or inspect.isfunction(value))
                    and not value.__module__ == "__main__"
                    and not typic.isbuiltintype(value)
                ):
                    imports.l(
                        f"{Keyword.FRM} {value.__module__} {Keyword.IMP} {value.__name__} as {key}",
                        level=0,
                    )
                else:
                    remains[key] = value
        return imports, remains

    @staticmethod
    def globals(namespace: dict) -> Tuple[Block, dict]:
        remains = {}
        with Block() as g:
            for key, value in namespace.items():
                if isinstance(value, tuple(typic.BUILTIN_TYPES)):
                    g.l(f"{key} = {value}", level=0)
                else:
                    # Can't set values for custom types/classes when writing a file
                    remains[key] = value
        return g, remains

    def render(self) -> Tuple[str, dict]:
        imports, remains = self.imports(self.namespace)
        globs, remains = self.globals(remains)
        return "".join((imports.render(), globs.render(), self.body.render())), remains

    def compile(self, ns: dict):
        code, namespace = self.render()
        ns.update(namespace)
        exec(code, ns, ns)

    def write(self, path: pathlib.Path):
        code, namespace = self.render()
        path.write_text(code)
        return code



# flake8: noqa
# pragma: nocover
from __future__ import annotations

import sys
from typing import (
    TYPE_CHECKING,
    Any,
    _SpecialForm as SpecialForm,  # type: ignore
    TypeVar,
    Optional,
    Callable,
)

try:
    from typing import Final, TypedDict, Literal, Protocol, TypeGuard, get_origin, get_args  # type: ignore
except ImportError:  # pragma: nocover
    from typing_extensions import Final, TypedDict, Literal, Protocol, TypeGuard, get_origin, get_args  # type: ignore
try:
    from typing import ForwardRef  # type: ignore
except ImportError:  # pragma: nocover
    from typing import _ForwardRef as ForwardRef  # type: ignore
try:
    from sqlalchemy import MetaData as SQLAMetaData  # type: ignore
except ImportError:  # pragma: nocover

    class SQLAMetaData:  # type: ignore
        ...


try:
    from sqlalchemy.orm import registry as sqla_registry  # type: ignore
except ImportError:  # pragma: nocover

    class sqla_registry:  # type: ignore
        ...


if sys.version_info < (3, 7):  # pragma: nocover
    if TYPE_CHECKING:

        class ForwardRef:
            __forward_arg__: str

            def _eval_type(self, globalns: Any, localns: Any) -> Any:
                pass

    else:
        from typing import _ForwardRef as ForwardRef

    def evaluate_forwardref(
        type_: ForwardRef,
        globalns: Any,
        localns: Any,
        recursive_guard: set = None,
    ) -> Any:
        return type_._eval_type(globalns, localns)


elif sys.version_info >= (3, 9):  # pragma: nocover
    from typing import ForwardRef

    def evaluate_forwardref(
        type_: ForwardRef,
        globalns: Any,
        localns: Any,
        recursive_guard: set = None,
    ) -> Any:
        recursive_guard = recursive_guard or set()
        return type_._evaluate(globalns, localns, recursive_guard)  # type: ignore


else:  # pragma: nocover
    from typing import ForwardRef  # type: ignore

    def evaluate_forwardref(
        type_: ForwardRef,
        globalns: Any,
        localns: Any,
        recursive_guard: set = None,
    ) -> Any:
        return type_._evaluate(globalns, localns)  # type: ignore


if TYPE_CHECKING:
    F = TypeVar("F", bound=Callable)

    def lru_cache(
        maxsize: Optional[int] = 128, typed: bool = False
    ) -> Callable[[F], F]:
        pass


else:
    from functools import lru_cache


if sys.version_info >= (3, 10):  # pragma: nocover
    DATACLASS_KW_ONLY = DATACLASS_MATCH_ARGS = DATACLASS_NATIVE_SLOTS = True
    from dataclasses import KW_ONLY  # type: ignore

else:
    DATACLASS_KW_ONLY = DATACLASS_MATCH_ARGS = DATACLASS_NATIVE_SLOTS = False

    class _KW_ONLY_TYPE:
        pass

    KW_ONLY = _KW_ONLY_TYPE()

__all__ = (
    "Final",
    "TypedDict",
    "Literal",
    "Protocol",
    "TypeGuard",
    "ForwardRef",
    "SQLAMetaData",
    "sqla_registry",
    "evaluate_forwardref",
    "lru_cache",
    "DATACLASS_KW_ONLY",
    "DATACLASS_MATCH_ARGS",
    "DATACLASS_NATIVE_SLOTS",
    "KW_ONLY",
)



from __future__ import annotations

import dataclasses
import uuid
from types import MappingProxyType
from typing import (
    Type,
    ClassVar,
    Mapping,
    Dict,
    Any,
    Union,
    FrozenSet,
    Pattern,
    Callable,
    Optional,
    Tuple,
    Collection,
    List,
    Hashable,
    TYPE_CHECKING,
)

from typic import gen, util
from typic.types.frozendict import FrozenDict, freeze
from .common import (
    BaseConstraints,
    ContextT,
    AssertionsT,
    VT,
    InstanceCheck,
    ConstraintsProtocolT,
)
from .error import ConstraintSyntaxError

if TYPE_CHECKING:  # pragma: nocover
    from typic.constraints.factory import ConstraintsT  # noqa: F401


def validate_pattern_constraints(
    constraints: Dict[Pattern, ConstraintsT], key: str, val: VT
) -> VT:
    for pattern, const in constraints.items():
        if pattern.match(key):
            val = const.validate(val)
    return val


MappedItemConstraints = Dict[Type, BaseConstraints]
ItemValidator = Union[
    Callable[[BaseConstraints, VT], VT], Callable[[MappedItemConstraints, VT], VT]
]


@util.slotted(dict=False)
@dataclasses.dataclass
class ItemValidatorNames:
    item_validators_name: str
    vals_validator_name: str
    keys_validator_name: str
    patterns_validators_name: str


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class MappingConstraints(BaseConstraints):
    type: ClassVar = Mapping
    min_items: Optional[int] = None
    """The minimum number of items which must be present in this mapping."""
    max_items: Optional[int] = None
    """The maximum number of items which may be present in this mapping."""
    required_keys: FrozenSet[str] = dataclasses.field(default_factory=frozenset)
    """A frozenset of keys which must be present in the mapping."""
    key_pattern: Optional[Pattern] = None
    """A regex pattern for which all keys must match."""
    items: Optional[FrozenDict[Hashable, ConstraintsProtocolT]] = None
    """A mapping of constraints associated to specific keys."""
    patterns: Optional[FrozenDict[Pattern, ConstraintsProtocolT]] = None
    """A mapping of constraints associated to any key which match the regex pattern."""
    values: Optional[ConstraintsProtocolT] = None
    """Whether values not defined as required are allowed.

    May be a boolean, or more constraints which are applied to all additional values.
    """
    keys: Optional[ConstraintsProtocolT] = None
    """Constraints to apply to any additional keys not explicitly defined."""
    key_dependencies: Optional[FrozenDict[str, KeyDependencyT]] = None
    """A mapping of keys and their dependent restrictions if they are present."""
    total: Optional[bool] = False
    """Whether to consider this schema as the 'total' representation.

    - If a mapping is `total=True`, no additional keys/values are allowed and cannot be
      defined.
    - Conversely, if a mapping is `total=False`, `required_keys` cannot not be
      defined.
    """
    X = "x"
    Y = "y"
    RETX = "retx"
    RETY = "rety"
    RETVAL = "retval"

    def _set_item_validator_loop_line(
        self, loop: gen.Block, func_name: str
    ) -> ContextT:
        names = ItemValidatorNames(
            item_validators_name=f"{func_name}_items",
            vals_validator_name=f"{func_name}_vals",
            keys_validator_name=f"{func_name}_keys",
            patterns_validators_name=f"{func_name}_patterns",
        )
        ctx: Dict[str, Any] = {}
        x = self.X
        y = self.Y
        field = f"_lazy_repr({self.FNAME}, {self.X})"
        if self.values:
            y = f"{names.vals_validator_name}({y}, field={field})"
            ctx[names.vals_validator_name] = self.values.validate
        if self.keys:
            x = f"{names.keys_validator_name}({self.X})"
            ctx[names.keys_validator_name] = self.keys.validate
        if self.patterns:
            y = (
                "validate_pattern_constraints"
                f"({names.patterns_validators_name}, {self.X}, {y})"
            )
            ctx[names.patterns_validators_name] = self.patterns
            ctx["validate_pattern_constraints"] = validate_pattern_constraints
        if self.items:
            ctx.update(
                {
                    names.item_validators_name: MappingProxyType(
                        {x: y.validate for x, y in self.items.items()}  # type: ignore
                    )
                }
            )
            y = (
                f"{names.item_validators_name}[{self.X}]({y}, field={field}) "
                f"if {self.X} in {names.item_validators_name} else {self.Y}"
            )

        loop.l(f"{x}: {y}")
        return ctx

    def _set_item_validator_pattern_constraints(self, loop: gen.Block, func_name: str):
        # Item constraints based upon key-pattern
        pattern_constr_name = f"{func_name}_pattern_constraints"
        if self.patterns:
            loop.l(
                f"{self.RETY} = "
                f"validate_pattern_constraints"
                f"({pattern_constr_name}, {self.X}, {self.Y})",
                level=None,
                **{
                    "validate_pattern_constraints": validate_pattern_constraints,
                    pattern_constr_name: self.patterns,
                },
            )
        # Required key pattern
        if self.key_pattern:
            key_pattern_name = f"{func_name}_key_pattern"
            loop.l(
                f"valid = bool({key_pattern_name}.match({self.X}))",
                level=None,
                **{key_pattern_name: self.key_pattern},
            )
            with loop.b("if not valid:") as b:
                b.l("break")

    def _build_item_validator(self, func: gen.Block) -> Optional[ContextT]:
        if any(
            (
                self.items,
                self.patterns,
                self.key_pattern,
                self.keys,
                self.values,
            )
        ):
            with func.b(f"{self.VALUE} = {{") as loop:
                item_context = self._set_item_validator_loop_line(loop, func.name)
                loop.l(f"for {self.X}, {self.Y} in {self.VALUE}.items()")
            func.l("}")
            if self.key_pattern:
                key_pattern_name = f"{func.name}_key_pattern"
                with func.b(
                    f"if any((not {key_pattern_name}.match({self.X}) "
                    f"for {self.X} in {self.VALUE})):"
                ) as b:
                    b.l(f"return False, {self.VALUE}")
                    item_context[key_pattern_name] = self.key_pattern
            return item_context
        return None

    def _get_key_dependencies(self, assertions: AssertionsT, context: ContextT):
        for key, dep in self.key_dependencies.items():  # type: ignore
            # If it's a collection, then we're just checking if another set of keys exist.
            if isinstance(dep, Collection) and not isinstance(
                dep, (Mapping, str, bytes)
            ):
                line = (
                    f"{{*{self.VALUE}.keys()}}.issuperset({set(dep)})"
                    f"if {key!r} in {self.VALUE} else True"
                )
            # If it's an instance of mapping constraints,
            # then we validate the entire value against that constraint.
            elif isinstance(dep, MappingConstraints):
                name = f"__{key}_constr_{uuid.uuid4().int}"
                line = (
                    f"{name}.validate({self.VALUE}, field=field) "
                    f"if {key!r} in {self.VALUE} else True"
                )
                context[name] = dep
            # Fail loud and make the user fix that shit.
            else:
                raise ConstraintSyntaxError(
                    f"Got an unsupported dependency in {self!r} for key {key!r}: {dep!r}"
                )

            assertions.append(line)

    def _get_assertions(self) -> AssertionsT:
        assertions: List[str] = []
        if self.min_items is not None:
            assertions.append(f"size >= {self.min_items}")
        if self.max_items is not None:
            assertions.append(f"size <= {self.max_items}")
        if self.required_keys:
            assertions.append("valkeys.issuperset(required)")
        defined_keys = (self.required_keys or set()) | (self.items or {}).keys()
        if defined_keys:
            assertions.append(
                f"valkeys.{'issubset' if self.total else 'issuperset'}(defined)"
            )
        return assertions

    def _build_assertions(self, func: gen.Block, assertions: AssertionsT):
        if assertions:
            if (self.max_items, self.min_items) != (None, None):
                func.l(f"size = len({self.VALUE})")
            BaseConstraints._build_assertions(self, func, assertions)

    def _check_syntax(self):
        if self.total and (self.keys or self.values):
            raise ConstraintSyntaxError(
                f"A mapping may not be considered 'total' and allow additional "
                f"keys/values: {self}"
            )

    def _build_validator(
        self, func: gen.Block, context: Dict[str, Any], assertions: AssertionsT
    ) -> ContextT:
        if self.key_dependencies:
            self._get_key_dependencies(assertions, context)
        _lazy_repr = (
            util.collectionrepr if issubclass(self.type, Mapping) else util.joinedrepr
        )
        context.update(Mapping=Mapping, _lazy_repr=_lazy_repr)
        if self.required_keys:
            context["required"] = self.required_keys
        defined_keys = (self.required_keys or set()) | (self.items or {}).keys()
        if defined_keys:
            context["defined"] = frozenset(defined_keys)
        if not issubclass(self.type, Mapping):
            with func.b(f"if not isinstance({self.VALUE}, Mapping):") as b:
                b.l(f"return False, {self.VALUE}")
        func.l(f"valkeys = {{*{self.VALUE}}}")
        context = BaseConstraints._build_validator(
            self, func=func, context=context, assertions=assertions
        )
        items_context = self._build_item_validator(func)
        if items_context:
            context.update(items_context)
        return context

    def for_schema(self, *, with_type: bool = False) -> dict:
        props = (
            freeze({x: y.for_schema() for x, y in self.items.items()})
            if self.items
            else None
        )
        schema: Dict[str, Any] = dict(
            title=self.name,
            minProperties=self.min_items,
            maxProperties=self.max_items,
            required=tuple(self.required_keys) or None,
            properties=props,
            propertyNames=(
                {"pattern": self.key_pattern.pattern} if self.key_pattern else None
            ),
        )
        if with_type:
            schema["type"] = "object"
        return {x: y for x, y in schema.items() if y is not None}


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class DictConstraints(MappingConstraints):
    type: ClassVar[Type[dict]] = dict


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class ObjectConstraints(MappingConstraints):
    type: Type = dataclasses.field(default=object)  # type: ignore
    instancecheck: ClassVar[InstanceCheck] = InstanceCheck.IS
    total: bool = True
    coerce: bool = True


@util.slotted
@dataclasses.dataclass(frozen=True, repr=False)
class TypedDictConstraints(ObjectConstraints):
    instancecheck: ClassVar[InstanceCheck] = InstanceCheck.NOT
    ttype: Type = dict

    __ignore_repr__ = frozenset(("ttype",)) | ObjectConstraints.__ignore_repr__

    @util.cached_property
    def type_name(self):
        return util.get_name(self.ttype)


KeyDependencyT = Union[Tuple[str], MappingConstraints]
"""A 'key dependency' defines constraints which are applied *only* if a key is present.

This can be either a tuple of dependent keys, or an additional MappingConstraints, which
is treated as a sub-schema to the parent MappingConstraints.
"""



from typing import Callable, Any, AnyStr

dumps: Callable[..., AnyStr]
loads: Callable[..., Any]


try:
    import ujson

    dumps = ujson.dumps  # type: ignore
    loads = ujson.loads

except (ImportError, ModuleNotFoundError):  # pragma: nocover
    import json

    dumps = json.dumps  # type: ignore
    loads = json.loads
